from __future__ import annotations

__all__ = [
    "Video",
]

import struct
from copy import copy, deepcopy
from mimetypes import guess_type
from os import makedirs
from os.path import abspath, dirname, isfile, join
from typing import Callable, Generator, Iterable, List, Sequence, Tuple

import cv2  # TODO: Compile OpenCV with GPU
import numpy as np
import torch
from numpy.typing import NDArray
from tqdm.auto import tqdm

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import SignFormats
from sign_language_translator.utils import in_jupyter_notebook
from sign_language_translator.vision.sign.sign import Sign
from sign_language_translator.vision.utils import (
    _normalize_args_index_and_timestamp,
    _validate_and_normalize_slices,
)
from sign_language_translator.vision.video.display import VideoDisplay
from sign_language_translator.vision.video.video_iterators import (
    IterableFrames,
    SequenceFrames,
    VideoCaptureFrames,
    VideoFrames,
)


class Video(Sign, VideoFrames):
    def __init__(
        self,
        sign: str
        | Sequence[NDArray[np.uint8] | torch.Tensor]
        | NDArray[np.uint8]
        | torch.Tensor
        | Generator[NDArray[np.uint8], None, None],
        **kwargs,
    ) -> None:
        self._path: str | None = None

        # ToDo: use list of sources instead of linked list of videos.
        # source
        self.source: VideoFrames
        self.__next: Video | None = None

        # play
        self._source_start_index: int = 0
        self._source_end_index: int
        # TODO: handle len & end_index for ||step|| > 1
        self._default_step_size: int = 1
        self.transformations: List[Callable] = []

        # properties
        self.fps: float = 30.0
        self.fourcc: int = cv2.VideoWriter_fourcc(*"XVID")  # type: ignore
        self._height: int
        self._width: int
        self._n_channels: int

        self.__initialize_from_arguments(sign, **kwargs)
        # TODO: from URL
        self._source_end_index = len(self.source) - 1
        self.__update_shape_components()

        # TODO: self.gpu_frame = cv2.cuda_GpuMat()
        self.__idx = 0

    @staticmethod
    def name() -> str:
        return SignFormats.VIDEO.value

    # ========================= #
    #    Iterate / get frame    #
    # ========================= #

    def get_frame(
        self, timestamp: float | None = None, index: int | None = None
    ) -> NDArray[np.uint8]:
        # allow negative indexing
        if index and index < 0:
            index = index + len(self)
        if timestamp and timestamp < 0:
            timestamp = timestamp + self.duration

        # validate arguments
        timestamp, _index = _normalize_args_index_and_timestamp(
            timestamp,
            index,
            self.duration if timestamp or self.fps else float("inf"),
            len(self) - 1,
        )

        # get frame
        node, relative_index = self.__get_node(_index)
        if node is not None:
            frame = node.source.get_frame(index=relative_index)
            for transformation in node.transformations:
                frame = transformation(frame)
            return frame

        # error
        raise IndexError(
            "Error reading frame at "
            + (f"{timestamp = }" if index is None else f"{index = }")
        )

    def trim(
        self,
        start_time: float | None = None,
        end_time: float | None = None,
        start_index: int | None = None,
        end_index: int | None = None,
    ) -> Video:
        # allow negative indexing
        if start_index and start_index < 0:
            start_index = start_index + len(self)
        if end_index and end_index < 0:
            end_index = end_index + len(self)
        if start_time and start_time < 0:
            start_time = start_time + self.duration
        if end_time and end_time < 0:
            end_time = end_time + self.duration

        # validate arguments
        start_time, start_index = _normalize_args_index_and_timestamp(
            start_time, start_index, self.duration or float("inf"), len(self) - 1
        )
        end_time, end_index = _normalize_args_index_and_timestamp(
            end_time, end_index, self.duration or float("inf"), len(self) - 1
        )
        if start_index > end_index:
            raise ValueError(
                f"invalid cut points. {start_index = } is not smaller than {end_index = }."
            )

        # get nodes to be trimmed
        start_node, relative_start_index = self.__get_node(start_index)
        end_node, relative_end_index = self.__get_node(end_index)
        if not end_node or not start_node:
            raise IndexError(
                f"Error trimming frames at {start_index = }, {end_index = }."
            )

        # trim
        remaining_nodes = end_node.__next
        end_node_source_end_index = end_node._source_end_index
        end_node.__next = None
        end_node._source_end_index = relative_end_index
        new = copy(start_node)
        end_node.__next = remaining_nodes
        end_node._source_end_index = end_node_source_end_index
        new._source_start_index = relative_start_index

        return new

    def iter_frames(
        self, start: int = 0, end: int | None = None, step: int | None = None
    ):
        for i in range(start, end or len(self), (step or 1) * self._default_step_size):
            yield self.get_frame(index=i)

    def __get_node(self, index: int) -> Tuple[Video | None, int]:
        """
        Find the node in the video linked list that contains the specified index/frame.

        Args:
            index (int): The index to search for within the video linked list.

        Returns:
            Video | None: The node containing the specified index or None if out of range.
        """

        # (0,(5,10),20) + (0,(6,9),10)
        # ....0,5.............6,9.....

        if index < 0:
            return None, -1
        elif index + self._source_start_index <= self._source_end_index:
            return self, index + self._source_start_index
        elif self.__next:
            return self.__next.__get_node(
                index - (self._source_end_index - self._source_start_index + 1)
            )
        else:
            return None, -1

    def __iter__(self):
        self.__idx = 0 if self._default_step_size > 0 else len(self) - 1
        return self

    def __next__(self) -> NDArray[np.uint8]:
        if 0 <= self.__idx < len(self):
            frame = self.get_frame(index=self.__idx)
            self.__idx += self._default_step_size
            return frame

        raise StopIteration

    def __getitem__(
        self, key: int | slice | Sequence[int | slice]
    ) -> Video | NDArray[np.uint8]:
        slices = _validate_and_normalize_slices(key, max_n_dims=4)
        if len(slices) > 4:
            raise ValueError(f"Expected at most 4 slices. Got {len(slices)}: {key}")
        if isinstance(key, int) or (
            isinstance(key, Sequence) and isinstance(key[0], int)
        ):
            return self.get_frame(index=slices[0].start)[slices[1:]]

        time_slice, frame_slices = slices[0], slices[1:]
        new = self.trim(
            start_index=time_slice.start or self._source_start_index,
            end_index=(
                time_slice.stop - 1
                if time_slice.stop is not None
                else self._source_end_index
            ),
        )

        # TODO: add support for step
        # every_node._default_step_size *= time_slice.step
        # every_node._source_end_index -= (node._source_end_index - node._source_end_index) % node._default_step_size

        new.transform(lambda frame: frame[frame_slices])

        return new

    # ========================= #
    #    Typecast / get data    #
    # ========================= #

    def numpy(self, *args, **kwargs):
        """
        Convert the video frames to a (4D) NumPy array.

        Args:
            *args: Positional arguments to be passed to `np.array()`.
            **kwargs: Keyword arguments to be passed to `np.array()`.

        Returns:
            np.ndarray: A NumPy array representing the video frames.
        """

        return np.array(list(self), *args, **kwargs)

    def __array__(self, *args, **kwargs):
        return self.numpy(*args, **kwargs)

    def torch(self, *args, **kwargs):
        """
        Convert the video frames to a (4D) PyTorch tensor.

        Args:
            *args: Positional arguments to be passed to `torch.tensor()`.
            **kwargs: Keyword arguments to be passed to `torch.tensor()`.

        Returns:
            torch.Tensor: A PyTorch tensor representing the video frames.
        """

        return torch.tensor(self.numpy(), *args, **kwargs)

    def __torch_function__(self, func, types, args=(), kwargs=None):
        if func is torch.Tensor or func is torch.tensor:
            dtype = kwargs.pop("dtype", torch.uint8)  # type: ignore
            return torch.tensor(self.numpy(), dtype=dtype, *args, **kwargs)  # type: ignore
        return NotImplemented

    # ========================== #
    #    Display / show frame    #
    # ========================== #

    def show(self, inline_player="html5" or "jshtml", **kwargs) -> None:
        if (
            in_jupyter_notebook()
            and self._path is not None
            and self.__next is None
            and inline_player == "html5"
            and not self.transformations
        ):
            VideoDisplay.display_ipython_video_in_jupyter(self._path)
        elif in_jupyter_notebook() and inline_player == "html5" and len(self) > 1:
            temp_filename = abspath(join(".", "__temp__.mp4"))
            self.save(temp_filename, overwrite=True, codec="avc1")
            VideoDisplay.display_ipython_video_in_jupyter(temp_filename)

        else:
            VideoDisplay.display_frames(
                self, fps=self.fps or 30.0, inline_player=inline_player
            )

    def show_frame(
        self, timestamp: float | None = None, index: int | None = None
    ) -> None:
        frame = self.get_frame(timestamp=timestamp, index=index)

        if frame is not None:
            VideoDisplay.display_frames([frame], inline_player="jshtml")

    def frames_grid(
        self, rows=2, columns=3, width: int | None = None, height: int | None = None
    ):
        grid = np.concatenate(
            [
                np.concatenate([self.get_frame(index=id) for id in ids_in_row], axis=1)
                for ids_in_row in np.linspace(0, len(self) - 1, rows * columns)
                .reshape(rows, columns)
                .round()
                .astype(int)
            ],
            axis=0,
        )

        if width or height:
            # calculate height or width of grid by maintaining aspect ratio of cell
            aspect_ratio = (self.width * columns) / (self.height * rows)

            if width and not height:
                height = int(width / aspect_ratio)
            elif height and not width:
                width = int(height * aspect_ratio)

            grid = cv2.resize(grid, (width, height))  # type: ignore

        return grid

    def show_frames_grid(
        self, rows=2, columns=3, width: int | None = 800, height: int | None = None
    ):
        grid = self.frames_grid(rows=rows, columns=columns, width=width, height=height)

        VideoDisplay.display_frames([grid], inline_player="jshtml")

    # ================= #
    #    Concatenate    #
    # ================= #

    @staticmethod
    def concatenate(objects: Iterable[Video]) -> Video:
        """
        Concatenate a sequence of Video objects into a single Video.

        Args:
            objects (Iterable[Video]): A sequence of Video objects to concatenate.

        Returns:
            Video: A new Video object that is a linked list of all the input videos.

        Raises:
            ValueError: If the input sequence of videos is empty.
        """

        video_chain = None
        for video in objects:
            if video_chain is None:
                video_chain = deepcopy(video)
            else:
                video_chain.append(video)

        if video_chain is None:
            raise ValueError("Cannot concatenate an empty sequence of videos.")

        return video_chain

    @staticmethod
    def stack(videos: List[Video], dim=1):
        """
        Stack a list of Video objects along a specified dimension (0=time, 1=height, 2=width, 3=channels).

        Args:
            videos (List[Video]): A list of Video objects to stack.
            dim (int, optional): The dimension along which to stack the videos. Defaults to 1.

        Returns:
            Video: A new Video object resulting from stacking the input videos.
        """

        if dim == 0:
            return Video.concatenate(videos)

        videos = list(videos)

        def stacked_frames(videos: List[Video]):
            for i in range(max(len(v) for v in videos)):
                yield np.concatenate(
                    [v.get_frame(index=i) for v in videos], axis=dim - 1
                )

        return Video(stacked_frames(videos), fps=videos[0].fps)

    def append(self, other: Video):
        """
        Append another Video object to the end of the current Video.

        Args:
            other (Video): The Video object to append.

        Returns:
            None
        """

        # prevent cycle in the video linked list
        other = copy(other)

        if self.__next is None:
            self.__next = other
        else:
            self.__next.append(other)

    def __add__(self, other: Video):
        new = copy(self)
        new.append(copy(other))
        return new

    def __copy__(self):
        new = self.__class__.__new__(self.__class__)
        new.__dict__.update({k: copy(v) for k, v in self.__dict__.items()})
        new._source_end_index = self._source_end_index

        return new

    # =============== #
    #    Transform    #
    # =============== #

    def transform(
        self, transformation: Callable[[NDArray[np.uint8]], NDArray[np.uint8]]
    ):
        """
        Apply a transformation function to the individual frames of this object (in place).

        Args:
            transformation (Callable[[NDArray[np.uint8]], NDArray[np.uint8]]): The transformation function to be applied that must map 1 numpy array to another numpy array.

        Raises:
            ValueError: If the transformation is not callable.

        Returns:
            None
        """

        if not isinstance(transformation, (Callable,)):
            raise ValueError(
                f"Cannot apply this transformation because it is not callable. {transformation = }"
            )

        self.transformations.append(transformation)
        if self.__next:
            self.__next.transform(transformation)

        self.__update_shape_components()

    # ================ #
    #    Dimensions    #
    # ================ #

    def __len__(self) -> int:
        # TODO: handle steps e.g. video[::3, ...]
        # todo: fix len when transformation is applied
        return (self._source_end_index - self._source_start_index + 1) + len(
            self.__next or ()
        )

    @property
    def shape(self) -> Tuple[int, int, int, int]:
        """Tuple of array dimensions (n_frames, height, width, n_channels)."""
        return len(self), self.height, self.width, self.n_channels

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width

    @property
    def n_channels(self) -> int:
        return self._n_channels

    @property
    def duration(self) -> float:
        """total time that the frames would take to play in a sequence. depends on fps."""
        if self.fps is None:
            # TODO: remove this
            raise RuntimeError("FPS is not set. Can not get duration without fps.")

        return len(self) / self.fps if self.fps else float("inf")

    @property
    def length(self) -> int:
        return len(self)

    # ================== #
    #    Save to disk    #
    # ================== #

    @staticmethod
    def save_(  # TODO: overload save to be static and instance method simultaneously
        frames_iterable: Iterable[NDArray[np.uint8]],
        path: str,
        overwrite=False,
        height: int | None = None,
        width: int | None = None,
        fps: float = 30.0,
        codec: str = "XVID",
        progress_bar=True,
        leave=False,
        total_frames: int | None = None,
        **kwargs,
    ) -> None:
        path = abspath(path)
        if isfile(path) and not overwrite:
            raise FileExistsError(
                f"File '{path}' already exists. Use overwrite=True to overwrite."
            )
        makedirs(dirname(path), exist_ok=True)

        _frames_iterable = iter(
            frames_iterable
            if not progress_bar
            else tqdm(
                frames_iterable,
                total=getattr(frames_iterable, "__len__", None) or total_frames,
                leave=leave,
            )
        )
        first_frame = next(_frames_iterable)
        height = height or first_frame.shape[0]
        width = width or first_frame.shape[1]

        # TODO: Compile opencv with GPU support and write to disk form CUDA.
        writer = cv2.VideoWriter(
            path, cv2.VideoWriter_fourcc(*codec), fps, (width, height)  # type: ignore
        )

        try:
            writer.write(cv2.cvtColor(first_frame, cv2.COLOR_BGR2RGB))
            for frame in _frames_iterable:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                writer.write(frame)
        finally:
            writer.release()

    def save(
        self,
        path: str,
        overwrite=False,
        fps: float | None = None,
        codec: str | None = None,
        progress_bar=True,
        leave=False,
        **kwargs,
    ) -> None:
        """Save the video frames to a file.

        Args:
            path (str): The path to the output file.
            fps (float | None, optional): The frames per second of the output video. If None, it will use the fps of the video source or 30 if that is not set either. Defaults to None.
            codec (str | None, optional): The codec used for the output video e.g ["xvid", "h264", "hvc1"]. If None, it will use the source video codec or "XVID" if that is not set. Defaults to None.
            **kwargs: Additional keyword arguments. (not used yet.)
        """

        Video.save_(
            self.iter_frames(),
            path,
            overwrite=overwrite,
            width=self.width,
            height=self.height,
            fps=fps or self.fps or 30,
            codec=codec or self.codec or "XVID",
            progress_bar=progress_bar,
            leave=leave,
            total_frames=len(self),
            **kwargs,
        )

    def save_frame(
        self,
        path: str,
        timestamp: float | None = None,
        index: int | None = None,
        overwrite=False,
    ) -> None:
        """
        Saves a single frame from the video object to the specified path.

        Args:
            path (str): The path where the frame will be saved.
            timestamp (float | None, optional): The timestamp of the frame to be saved. If None, `index` will be used. Defaults to None.
            index (int | None, optional): The index of the frame to be saved. If None, `timestamp` will be used. Defaults to None.
        Raises:
            ValueError: If both or neither `timestamp` and `index` are provided or If the specified timestamp or index is out of range.
            RuntimeError: If there is an error reading the frame.
        """

        path = abspath(path)
        if isfile(path) and not overwrite:
            raise FileExistsError(
                f"File '{path}' already exists. Use overwrite=True to overwrite."
            )
        makedirs(dirname(path), exist_ok=True)

        frame = self.get_frame(timestamp=timestamp, index=index)
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        cv2.imwrite(path, frame)

    def save_frames_grid(
        self,
        path: str,
        rows: int = 2,
        columns: int = 3,
        width: int | None = 1024,
        height: int | None = None,
        overwrite=False,
    ) -> None:
        path = abspath(path)
        if isfile(path) and not overwrite:
            raise FileExistsError(
                f"File '{path}' already exists. Use overwrite=True to overwrite."
            )
        makedirs(dirname(path), exist_ok=True)

        grid = self.frames_grid(rows=rows, columns=columns, width=width, height=height)
        grid = cv2.cvtColor(grid, cv2.COLOR_RGB2BGR)
        cv2.imwrite(path, grid)

    # ============================= #
    #    Initialize Video object    #
    # ============================= #

    @staticmethod
    def load(path: str, **kwargs) -> Video:
        """
        Load a video from the specified path.

        Args:
            path (str): The path to the video file.
            **kwargs: Additional keyword arguments to be passed to the Video constructor.

        Returns:
            Video: The loaded video object.
        """
        return Video(path, **kwargs)

    def __initialize_from_arguments(self, sign, **kwargs):
        if isinstance(sign, str):
            if isfile(sign) and not kwargs.get("is_asset", False):
                self._from_path(sign, **kwargs)
            elif Assets.get_ids(sign):
                Assets.download(sign, leave=False, overwrite=False)
                self._from_path(Assets.get_path(sign)[0], **kwargs)
            else:
                raise ValueError(
                    f"Invalid argument: {sign}. provide path to a video or relative path to a video resource from dataset."
                )

        elif isinstance(sign, (np.ndarray, torch.Tensor, Sequence)):
            self._from_frames(sign, **kwargs)  # type: ignore
        elif isinstance(sign, Iterable):
            self._from_iterable(
                sign,
                kwargs["total_frames"],
                **{k: v for k, v in kwargs.items() if k not in ("total_frames",)},
            )

        else:
            raise ValueError(
                f"Invalid argument: {sign}. provide either a path to a video or a sequence of frames."
            )

    def _from_path(self, path: str, **kwargs) -> None:
        self._path = abspath(path)
        file_type = str(guess_type(path)[0])

        if file_type.startswith("image"):
            self._from_frames(
                [cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)],  # type: ignore
                fps=kwargs.get("fps", 1),
                **kwargs,
            )

        elif file_type.startswith("video"):
            self.source = VideoCaptureFrames(self._path, **kwargs)
            self.__extract_properties_from_source(self.source)

        else:
            raise ValueError(f"unknown file type: {file_type}")

    def __extract_properties_from_source(self, source: VideoFrames) -> None:
        self.fps = getattr(source, "fps", 30.0)
        self.fourcc = getattr(source, "fourcc", cv2.VideoWriter_fourcc(*"XVID"))  # type: ignore

    @property
    def codec(self) -> str:
        return struct.pack("<I", self.fourcc).decode("utf-8")

    def __update_shape_components(self) -> None:
        shape = self.get_frame(0).shape
        self._height = shape[0]
        self._width = shape[1]
        self._n_channels = shape[2]

    def _from_data(self, sign_data: Sequence[NDArray[np.uint8]], **kwargs) -> None:
        if isinstance(sign_data, Sequence):
            self._from_frames(sign_data, **kwargs)
        else:
            self._from_iterable(sign_data, kwargs["total_frames"], **kwargs)

    def _from_frames(
        self, frames: Sequence[NDArray[np.uint8]], fps: float = 30.0, **kwargs
    ):
        # validation
        if len(frames) <= 0:
            raise ValueError("Can not initialize from None or empty frames argument.")
        if frames[0].ndim != 3:
            raise ValueError("Each frame must be 3D (Height, Width, Channels).")
        if isinstance(frames, torch.Tensor):
            frames = frames.numpy().astype(np.uint8)
        elif isinstance(frames[0], torch.Tensor):
            frames = torch.Tensor(frames).numpy().astype(np.uint8)

        # source
        self.source = SequenceFrames(frames, fps=fps)
        self.__extract_properties_from_source(self.source)

    def _from_iterable(self, frames_iterable: Iterable, total_frames: int, **kwargs):
        """load the frames from a read-once iterable such as a generator."""
        self.source = IterableFrames(frames_iterable, total_frames, **kwargs)
        self.__extract_properties_from_source(self.source)

    # ============================= #
    #    Cleaning / with _ as _:    #
    # ============================= #

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """Release the resources occupied by the object and reset the properties."""
        # source
        self.source.close()
        if self.__next is not None:
            self.__next.close()
            self.__next = None

        # play
        self._source_start_index = 0
        self._source_end_index = 0
        self.transformations = []

        # properties
        self._path = None

        self.is_closed = True



from __future__ import annotations

__all__ = [
    "Sign",
]

from abc import ABC, abstractmethod
from typing import Callable, Iterable

from numpy.typing import NDArray
from torch import Tensor


class Sign(ABC):
    def __init__(self, sign, *args, **kwargs) -> None:
        """initialize the sign"""

    @staticmethod
    @abstractmethod
    def name() -> str:
        """return the name of the sign format"""

    @abstractmethod
    def show(self, **kwargs) -> None:
        """display the sign"""

    @abstractmethod
    def _from_path(self, path: str):
        """load the sign from a path"""

    # _from_dataset

    @abstractmethod
    def _from_data(self, sign_data):
        """load the sign from an object"""

    @staticmethod
    @abstractmethod
    def concatenate(objects: Iterable[Sign]) -> Sign:
        """concatenate multiple signs"""

    @abstractmethod
    def numpy(self, *args, **kwargs) -> NDArray:
        """return the sign as a numpy array"""

    @abstractmethod
    def torch(self, *args, **kwargs) -> Tensor:
        """return the sign as a torch tensor"""

    @abstractmethod
    def transform(self, transformation: Callable):
        """apply some transformation to the sign to change its appearance"""

    @abstractmethod
    def save(self, path: str, *args, **kwargs) -> None:
        """save the sign to a path"""



from abc import ABC, abstractmethod
from typing import Any, Iterable, Union


class TextToSignModel(ABC):
    @property
    @abstractmethod
    def sign_language(self):
        """The target sign language of the model."""

    @property
    @abstractmethod
    def text_language(self):
        "The source text language of the model."

    @property
    @abstractmethod
    def sign_format(self):
        """The format of the sign language (e.g. slt.Vision.sign.sign.Sign)."""

    @abstractmethod
    def translate(self, text: Union[str, Iterable[str]], *args, **kwargs) -> Any:  # -> VideoFeatures
        """Translate the text to sign language."""

    def __call__(self, text: Union[str, Iterable[str]], *args, **kwargs) -> Any:
        return self.translate(text, *args, **kwargs)



"""Custom Layers for Decoder-only Transformers

This module contains custom layers used in the Transformer Decoder architecture.

Classes:
    FeedForward(torch.nn.Module): Implements a simple feedforward neural network module
        with one hidden layer.
    CausalMultiHeadSelfAttention(torch.nn.Module): Implements the causal multi-head self-attention
        mechanism used in transformer decoders.
    DecoderBlock(torch.nn.Module): Implements a single transformer decoder block with
        multi-head self-attention and feedforward neural network layers but no cross-attention.

Example:
    >>> import torch
    >>> from sign_language_translator.models.language_models.transformer_language_model.layers import FeedForward, DecoderBlock, CausalMultiHeadSelfAttention

    >>> model = FeedForward(n_embed=256, hidden_size=512, dropout=0.2, activation='relu')
    >>> input_tensor = torch.randn(32, 256)
    >>> output_tensor = model(input_tensor)

    >>> decoder_block = DecoderBlock(n_embed=256, hidden_size=512, n_heads=8, max_seq_len=32, dropout=0.2, activation='relu')
    >>> input_tensor = torch.randn(16, 32, 256)
    >>> output_tensor = decoder_block(input_tensor)

    >>> attention_layer = CausalMultiHeadSelfAttention(n_heads=8, embed_size=256, dropout=0.2)
    >>> input_tensor = torch.randn(16, 32, 256)
    >>> output_tensor = attention_layer(input_tensor)
"""

import torch


class CausalMultiHeadSelfAttention(torch.nn.Module):
    """
    Causal Multi-Head Self-Attention Module.

    This class implements the causal multi-head self-attention mechanism. It takes an input tensor of shape
    (batch_size, seq_len, embed_size) and applies causal attention, where each token can attend only to
    itself and the previous tokens in the sequence. The input tensor is transformed into queries, keys,
    and values, and then passed through the scaled dot-product attention mechanism. The final output tensor
    is obtained by concatenating the heads and applying a linear projection with dropout.

    Parameters:
        n_heads (int): The number of attention heads.
        embed_size (int): The size of the input embedding dimension. Must be divisible by n_heads.
        dropout (float, optional): The dropout probability applied in the attention and projection layers. Default is 0.25.
        max_seq_len (int, optional): The maximum input sequence length (used only in custom dot-product attention (pytorch<2.0.0)). Default is 64.
        attention_bias (bool, optional): If True, enables trainable bias parameter in the query, key & value layer. Default is False.

    Inputs:
        x (torch.Tensor): Input tensor of shape (batch_size, seq_len, embed_size).

    Returns:
        torch.Tensor: Output tensor of shape (batch_size, seq_len, embed_size).

    Example:
        >>> model = CausalMultiHeadSelfAttention(n_heads=8, embed_size=256, dropout=0.2)
        >>> input_tensor = torch.randn(16, 32, 256)
        >>> output_tensor = model(input_tensor)
    """

    def __init__(
        self,
        n_heads,
        embed_size,
        dropout=0.25,
        max_seq_len: int = 64,
        attention_bias=False,
    ):
        assert embed_size % n_heads == 0, "embed_dim must be divisible by num_heads"
        super().__init__()

        # save hyper parameters
        self.n_heads = n_heads
        self.embed_size = embed_size
        self.dropout_probability = dropout

        # causal attention mechanism
        self.use_flash_attention = hasattr(
            torch.nn.functional, "scaled_dot_product_attention"
        )
        if not self.use_flash_attention:
            # TODO: this layer should be agnostic of max_seq_len (auto grow)
            self.register_buffer(
                "fill_mask",
                torch.tril(torch.ones(max_seq_len, max_seq_len)).view(
                    -1, -1, max_seq_len, max_seq_len
                )
                == 0,
            )
        self.dot_product_scalar = (embed_size // n_heads) ** -0.5

        # learnable parameters
        self.all_queries_keys_values = torch.nn.Linear(
            embed_size, embed_size * 3, bias=attention_bias
        )
        self.attention_dropout = torch.nn.Dropout(dropout)
        self.projection = torch.nn.Linear(embed_size, embed_size)
        self.final_dropout = torch.nn.Dropout(dropout)

    def forward(self, x: torch.Tensor):
        """
        Forward pass of the Causal Multi-Head Self-Attention.

        This method applies the forward pass of the causal multi-head self-attention to the input tensor x.
        The input tensor is transformed into queries, keys, and values, which are then passed through the
        scaled dot-product attention mechanism. The final output tensor is obtained by concatenating the
        attention heads and applying a linear projection with dropout.

        Parameters:
            x (torch.Tensor): Input tensor of shape (batch_size, seq_len, embed_size).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, seq_len, embed_size).

        Example:
            >>> model = CausalMultiHeadSelfAttention(n_heads=8, embed_size=256, dropout=0.2)
            >>> input_tensor = torch.randn(16, 32, 256)
            >>> output_tensor = model.forward(input_tensor)
        """

        # ([batch,] seq_len, embed_size)
        seq_len = x.shape[-2]
        input_had_batch_dim = x.dim() == 2

        # get Q, K, V matrices ([batch,] seq_len, embed_size)
        x = self.all_queries_keys_values(x)  # ([batch,] seq_len, 3 * embed_size)
        querys, keys, values = x.split(self.embed_size, dim=-1)

        # break up embedding dimension ([batch,] seq_len,     n_heads, head_size)
        querys = querys.view(-1, seq_len, self.n_heads, self.embed_size // self.n_heads)
        keys = keys.view(-1, seq_len, self.n_heads, self.embed_size // self.n_heads)
        values = values.view(-1, seq_len, self.n_heads, self.embed_size // self.n_heads)

        # move n_heads dim to batch dim by swapping seq_len & n_heads using transpose
        querys = querys.transpose(-3, -2)  # (batch, n_head, seq_len, head_size)
        keys = keys.transpose(-3, -2)  #   (batch, n_head, seq_len, head_size)
        values = values.transpose(-3, -2)  # (batch, n_head, seq_len, head_size)

        if self.use_flash_attention:
            # GPU go Brrrrrr
            x = torch.nn.functional.scaled_dot_product_attention(
                querys,
                keys,
                values,
                attn_mask=None,
                dropout_p=self.dropout_probability if self.training else 0,
                is_causal=True,
            )  # (batch, n_head, seq_len, head_size)
        else:
            # Manual attention implementation
            # Q @ K = (B, nH, T, HS) @ (B, nH, HS, T) -> (B, nH, T, T)
            attentions = querys @ keys.transpose(-2, -1)
            # normalize the variance to 1
            attentions = attentions * self.dot_product_scalar
            # don't look at future tokens
            attentions = attentions.masked_fill(
                self.fill_mask[..., :seq_len, :seq_len], float("-inf")  # type: ignore
            )
            attentions = torch.nn.functional.softmax(attentions, dim=-1)
            attentions = self.attention_dropout(attentions)

            x = attentions @ values  # (B, nH, T, T) @ (B, nH, T, HS) -> (B, nH, T, HS)

        # concatenate heads
        x = x.transpose(-3, -2).contiguous()  #     (batch, seq_len, n_head, head_size)
        x = x.view(-1, seq_len, self.embed_size)  # (batch, seq_len, embed_size)

        x = self.projection(x)
        x = self.final_dropout(x)

        if input_had_batch_dim and x.shape[0] == 1:
            x = x.squeeze(0)

        return x


class FeedForward(torch.nn.Module):
    """
    FeedForward Neural Network Module.

    This class implements a simple feedforward neural network module with one hidden layer.
    It takes an input tensor of shape (batch_size, n_embed) and applies a linear transformation,
    followed by an activation function (GELU or ReLU), and then another linear transformation
    with dropout applied. The final output tensor has the same shape as the input tensor.

    Parameters:
        n_embed (int): The size of the input feature dimension.
        hidden_size (int): The number of neurons in the hidden layer.
        dropout (float, optional): The dropout probability applied after the second linear layer.
            Default is 0.25.
        activation (str, optional): The activation function to be used. Supported values are 'gelu'
            for GELU activation and 'relu' for ReLU activation. Default is 'gelu'.

    Inputs:
        x (torch.Tensor): Input tensor of shape (batch_size, n_embed).

    Returns:
        torch.Tensor: Output tensor of shape (batch_size, n_embed).

    Example:
        >>> model = FeedForward(n_embed=256, hidden_size=512, dropout=0.2, activation='relu')
        >>> input_tensor = torch.randn(32, 256)
        >>> output_tensor = model(input_tensor)
    """

    def __init__(self, n_embed, hidden_size, dropout=0.25, activation="gelu"):
        super().__init__()
        self.fully_connected_1 = torch.nn.Linear(n_embed, hidden_size)
        self.activation = torch.nn.GELU() if activation == "gelu" else torch.nn.ReLU()
        self.fully_connected_2 = torch.nn.Linear(hidden_size, n_embed)
        self.dropout = torch.nn.Dropout(dropout)

    def forward(self, x):
        """
        Forward pass of the FeedForward neural network.

        This method applies the forward pass of the feedforward neural network to the input tensor x.
        The forward pass involves passing the input tensor through the hidden layer, followed by an
        activation function, and then through the output layer with dropout applied.

        Parameters:
            x (torch.Tensor): Input tensor of shape (batch_size, n_embed).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, n_embed).

        Example:
            >>> model = FeedForward(n_embed=256, hidden_size=512, dropout=0.2, activation='relu')
            >>> input_tensor = torch.randn(32, 256)
            >>> output_tensor = model.forward(input_tensor)
        """

        x = self.fully_connected_1(x)
        x = self.activation(x)
        x = self.fully_connected_2(x)
        x = self.dropout(x)
        return x


class DecoderBlock(torch.nn.Module):
    """
    Transformer Decoder Block Module.

    This class implements a single transformer decoder block, consisting of causal multi-head self-attention
    and feedforward neural network layers but no cross-attention. The input tensor x goes through the
    layer norm & attention mechanism and also forms a skip connection over them into another
    layer norm & feedforward neural network. The output also contains a residual connection over these two operations.

    Parameters:
        n_embed (int): The size of the input feature dimension and also the output feature dimension.
        hidden_size (int): The number of neurons in the feedforward neural network's hidden layer.
        n_heads (int): The number of attention heads for multi-head self-attention.
        max_seq_len (int): The maximum sequence length of the input tensor.
        dropout (float, optional): The dropout probability applied in both attention and feedforward
            layers. Default is 0.25.
        activation (str, optional): The activation function to be used in the feedforward neural network.
            Supported values are 'gelu' for GELU activation and 'relu' for ReLU activation. Default is 'gelu'.
        device (torch.device, optional): If provided, the attention and feedforward layers will be moved
            to this device. Default is None.

    Inputs:
        x (torch.Tensor): Input tensor of shape (batch_size, seq_len, n_embed).

    Returns:
        torch.Tensor: Output tensor of shape (batch_size, seq_len, n_embed).

    Example:
        >>> model = DecoderBlock(n_embed=256, hidden_size=512, n_heads=8, max_seq_len=32, dropout=0.2, activation='relu')
        >>> input_tensor = torch.randn(16, 32, 256)
        >>> output_tensor = model(input_tensor)
    """

    def __init__(
        self,
        n_embed,
        hidden_size,
        n_heads,
        max_seq_len,
        dropout=0.25,
        activation="gelu",
        attention_bias=False,
    ):
        super().__init__()
        self.attention = CausalMultiHeadSelfAttention(
            n_heads,
            n_embed,
            dropout=dropout,
            max_seq_len=max_seq_len,
            attention_bias=attention_bias,
        )
        self.feed_forward = FeedForward(
            n_embed, hidden_size, dropout=dropout, activation=activation
        )
        self.mha_layer_norm = torch.nn.LayerNorm(n_embed)
        self.ff_layer_norm = torch.nn.LayerNorm(n_embed)

    def forward(self, x):
        x = x + self.attention(self.mha_layer_norm(x))
        x = x + self.feed_forward(self.ff_layer_norm(x))
        return x


__all__ = [
    "DecoderBlock",
    "FeedForward",
    "CausalMultiHeadSelfAttention",
]



from typing import Iterable, List, Sequence, Type

import numpy as np
from numpy.typing import NDArray
from torch import Tensor
from torch import long as torch_long

__all__ = [
    "ArrayOps",
    "linear_interpolation",
]


class ArrayOps:
    @staticmethod
    def floor(array: NDArray | Tensor) -> NDArray | Tensor:
        if isinstance(array, np.ndarray):
            return np.floor(array)
        elif isinstance(array, Tensor):
            return array.floor()
        else:
            raise TypeError(f"Invalid type for flooring: {type(array)}")

    @staticmethod
    def ceil(array: NDArray | Tensor) -> NDArray | Tensor:
        if isinstance(array, np.ndarray):
            return np.ceil(array)
        elif isinstance(array, Tensor):
            return array.ceil()
        else:
            raise TypeError(f"Invalid type for ceiling: {type(array)}")

    @staticmethod
    def take(
        array: NDArray | Tensor, index: NDArray | Tensor | List, dim: int = 0
    ) -> NDArray | Tensor:
        if isinstance(array, np.ndarray):
            if not isinstance(index, np.ndarray):
                index = np.array(index)
            return np.take(array, index.astype(int), axis=dim)
        elif isinstance(array, Tensor):
            if not isinstance(index, Tensor):
                index = Tensor(index)
            return array.index_select(dim, index.type(torch_long))
        else:
            raise TypeError(f"Invalid type for taking: {type(array)}")

    @staticmethod
    def cast(
        x: NDArray | Tensor | List | Iterable, data_type: Type[np.ndarray | Tensor]
    ) -> NDArray | Tensor:
        if data_type == np.ndarray:
            return np.array(x)
        elif data_type == Tensor:
            return Tensor(x)
        else:
            raise TypeError(f"Invalid type for array casting: {data_type}")


def linear_interpolation(
    array: NDArray[np.float64] | Tensor | List,
    new_indexes: Sequence[int | float] | None = None,
    old_x: Sequence[int | float] | None = None,
    new_x: Sequence[int | float] | None = None,
    dim: int = 0,
) -> NDArray | Tensor:
    """
    Perform linear interpolation on a multidimensional array or tensor along a dimension.

    This function essentially connects all consecutive values in a multidimensional array with
    straight lines along a specified dimension, so that intermediate values can be calculated.
    It takes the input array, a set of new indexes or alternatively old & new coordinate axes,
    and a dimension along which to perform interpolation.

    Parameters:
        array (NDArray[np.float64] | Tensor): The input array or tensor to interpolate.
        new_indexes (Sequence[int | float] | None, optional): The new indexes at which to interpolate the data.
            If None, it infers new_indexes from `old_x` and `new_x`.
        old_x (Sequence[int | float] | None, optional): The old coordinate values corresponding to the data in `array`.
            If None, it uses `new_indexes` argument.
        new_x (Sequence[int | float] | None, optional): The new coordinate values corresponding to the .
            If None, it uses `new_indexes` argument.
        dim (int, optional): The dimension along which to perform interpolation. Default is 0.

    Returns:
        NDArray | Tensor: The result of linear interpolation along the specified dimension.

    Raises:
        ValueError: If both or neither of `new_indexes` and `old_x` & `new_x` are provided.
        If `new_indexes` is not 1 dimensional.

    Examples:
        .. code-block: python
            data = np.array([1, 2, 3, 5])
            new_indexes = np.array([1.5, 0.5, 2.5])
            interpolated_data = linear_interpolation(data, new_indexes)
            print(interpolated_data)
            # array([2.5, 1.5, 4. ])

            old_x = [0, 4, 4.5, 5]
            new_x = [0, 1, 2, 2.5, 3, 4, 5]
            interpolated_data = linear_interpolation(data, old_x=old_x, new_x=new_x)
            print(interpolated_data)
            # array([1.   , 1.25 , 1.5  , 1.625, 1.75 , 2.   , 5.   ])

    Note:
        This function supports both NumPy arrays and PyTorch tensors as input.

    """
    if isinstance(array, list):
        array = np.array(array)

    indexes_, dim = __validate_linear_interpolation_args(
        array, new_indexes, old_x, new_x, dim
    )

    # interpolate
    floor_values = ArrayOps.floor(indexes_)
    ceiling_values = ArrayOps.ceil(indexes_)
    fraction = (indexes_ - floor_values).reshape(
        [len(indexes_) if axis == dim else 1 for axis in range(array.ndim)]
    )

    interpolated = ArrayOps.take(array, floor_values, dim) * (1 - fraction)
    interpolated += ArrayOps.take(array, ceiling_values, dim) * fraction

    return interpolated


def __validate_linear_interpolation_args(array, new_indexes, old_x, new_x, dim):
    if (new_indexes is None and (old_x is None or new_x is None)) or (
        new_indexes is not None and (old_x is not None or new_x is not None)
    ):
        raise ValueError(
            "Either `new_indexes` or both `old_x` and `new_x` must be provided."
        )
    if new_indexes is None:
        new_indexes = np.interp(new_x, old_x, np.arange(len(old_x)))  # type: ignore

    if not isinstance(array, (np.ndarray, Tensor)):
        array = np.array(array)
    indexes_ = ArrayOps.cast(new_indexes, type(array))  # type: ignore

    if indexes_.ndim != 1:
        raise ValueError(
            f"Invalid indexes shape: {indexes_.shape}. Must be 1-dimensional."
        )
    if dim >= array.ndim:
        raise ValueError(f"Invalid dim: {dim}. Must be between 0 and {array.ndim-1}.")

    out_of_bounds = (indexes_ < 0) & (indexes_ > array.shape[dim] - 1)
    if out_of_bounds.any():
        raise ValueError(
            f"Invalid indexes for interpolation: {indexes_[out_of_bounds.tolist()]}. "
            f"Must be between >= 0 and <= {array.shape[0]-1}."
        )
    if dim < 0 or dim >= array.ndim:
        raise ValueError(f"Invalid dim: {dim}. Must be between 0 and {array.ndim-1}.")

    return indexes_, dim



"""
sign_language_translator.models.language_models.transformer_language_model
============================================================================

This module contains the implementation of the Transformer-based language model.

Components:
-----------
- TransformerLanguageModel: The main class representing the Transformer-based language model.
- layers: Custom implementation of neural network layers used in the Transformer model.
- train: Functions & Classes related to training the Transformer model including LM_Dataset and LM_Trainer.
"""

from sign_language_translator.models.language_models.transformer_language_model import (
    layers,
    train,
)
from sign_language_translator.models.language_models.transformer_language_model.model import (
    TransformerLanguageModel,
)

__all__ = [
    "TransformerLanguageModel",
    "layers",
    "train",
]



import re
from enum import EnumMeta
from random import choices
from typing import Any, Dict, List, Set

from tqdm.auto import tqdm

__all__ = [
    "PrintableEnumMeta",
    "ProgressStatusCallback",
    "extract_recursive",
    "in_jupyter_notebook",
    "is_regex",
    "sample_one_index",
    "search_in_values_to_retrieve_key",
]


def search_in_values_to_retrieve_key(
    code_name: str, class_to_codes: Dict[Any, Set[str]]
):
    # verify there is no repetition/reuse in language codes
    all_codes = [code for codes in class_to_codes.values() for code in codes]
    assert len(all_codes) == len(set(all_codes)), "code reused for multiple keys"

    for key, codes in class_to_codes.items():
        if code_name.lower() in codes:
            return key

    return None


def sample_one_index(weights: List[float], temperature: float = 1.0) -> int:
    """Select an item based on the given probability distribution.
    Returns the index of the selected item sampled from weighted random distribution.

    Args:
        weights (List[float]): the relative weights corresponding to each index.
        temperature (float): The temperature value for controlling the sampling behavior.
            High temperature means sampling probabilities are more uniform (says random things).
            Low temperature means that sampling probabilities are higher for bigger weights.
            Defaults to 1.0.

    Returns:
        int: The index of the chosen item.
    """

    return choices(
        range(len(weights)),
        weights=[w ** (1 / temperature) for w in weights],
        k=1,
    )[0]


def in_jupyter_notebook():
    """
    Checks if the code is running in a Jupyter notebook.

    Returns:
        bool: True if running in a Jupyter notebook, False otherwise.
    """

    try:
        from IPython import get_ipython  # type: ignore

        return "IPKernelApp" in get_ipython().config  # type: ignore
    except:  # pylint: disable = bare-except
        return False


def extract_recursive(data: Dict[str, Any], key: str) -> List[Any]:
    """
    Recursively extracts values associated with a specified key from a nested dictionary.

    Args:
        data (Dict[str, Any]): The input dictionary containing nested structures.
        key (str): The key for which values need to be extracted.

    Returns:
        List[Any]: A list containing all values associated with the specified key, extracted
                   recursively from the input dictionary.

    Examples:
        >>> data = {'a': 1, 'b': {'c': 2, 'd': {'e': 3, 'f': 4}}, 'g': [5, {'h': 6, 'e': 7}]}
        >>> extract_recursive(data, 'e')
        [3, 7]
        >>> extract_recursive(data, 'h')
        [6]
        >>> extract_recursive(data, 'x')
        []  # Key not found, returns an empty list.
    """

    extracted_values = []

    def extract(data: Dict, results: List):
        for k in data:
            if k == key:
                results.append(data[k])
            elif isinstance(data[k], dict):
                extract(data[k], results)
            elif isinstance(data[k], list):
                for item in data[k]:
                    if isinstance(item, dict):
                        extract(item, results)

    extract(data, extracted_values)

    return extracted_values


class PrintableEnumMeta(EnumMeta):
    """
    Metaclass for customizing the string representation of Enum classes.

    This metaclass overrides the __str__ & __repr__ method to provide a human-readable
    representation of Enum classes when they are printed. The generated string
    includes the class name and a formatted list of Enum members and their values.

    Example:

    .. code-block:: python

        class MyEnumClass(enum.Enum, metaclass=PrintableEnumMeta):
            MEMBER1 = "value_1"
            MEMBER2 = "value_2"

        print(MyEnumClass)

        # "MyEnumClass" enum class. Available values:
        # 1. MEMBER1 = value_1
        # 2. MEMBER2 = value_2
    """

    def __str__(cls):
        members = "\n".join(
            [
                f"{i + 1}. {member.name} = {member.value}"  # type:ignore
                for i, member in enumerate(cls)
            ]
        )
        return f'"{cls.__name__}" enum class. Available values:\n{members}'

    def __repr__(cls) -> str:
        return str(cls)


class ProgressStatusCallback:
    """
    A callback class to update a tqdm progress bar with custom status information.

    Args:
        tqdm_bar (tqdm): The tqdm progress bar to be updated.

    Attributes:
        tqdm_bar (tqdm): The tqdm progress bar associated with the callback.

    Methods:
        __call__(self, status: Dict[str, Any]):
            Update the tqdm progress bar with the provided status information.

    Example:
        # Instantiate a tqdm progress bar & callback
        progress_bar = tqdm(total=100, desc='Processing')
        callback = ProgressStatusCallback(tqdm_bar=progress_bar)

        # Update the progress bar inside some other function
        status_info = {'Epoch': 1, 'Loss': 0.123, 'Accuracy': 0.95}
        callback(status_info)
    """

    def __init__(self, tqdm_bar: tqdm):
        """
        Initialize the ProgressStatusCallback with a tqdm progress bar.

        Args:
            tqdm_bar (tqdm): The tqdm progress bar to be associated with the callback.
        """
        self.tqdm_bar = tqdm_bar

    def __call__(self, status: Dict[str, Any]):
        """
        Update the tqdm progress bar with the provided status information.

        Args:
            status (Dict[str, Any]): A dictionary containing custom status information.
                This information will be displayed as postfix on the tqdm progress bar.
        """
        self.tqdm_bar.set_postfix(status, refresh=True)


def is_regex(string: str | re.Pattern) -> bool:
    """Tests whether the argument is a regex or a regular string.

    Args:
        string (str | Pattern): The string to be tested.

    Returns:
        bool: whether the argument is a regex (True) or a regular string (False).
    """
    if isinstance(string, re.Pattern):
        return True

    if set("+*?|[]{}^$").intersection(set(string)):
        try:
            re.compile(string)
            return True
        except re.error:
            return False



"""This module contains classes to train transformer language models.

Classes:
    LM_Dataset(torch.utils.data.Dataset): subclass for language model. has a process function that can convert text file into a list of tensors.
    LM_Trainer: Trainer class for language model. runs training loop, prints metrics and makes model checkpoints.
"""

from __future__ import annotations

import os
from glob import glob
from time import time
from typing import TYPE_CHECKING, Callable, Dict, Iterable, List, Tuple

import numpy as np
import torch
from tqdm.auto import tqdm

from sign_language_translator.models.utils import (
    FullyLambdaLR,
    set_layers_trainability_,
)
from sign_language_translator.text.utils import make_ngrams

if TYPE_CHECKING:
    from sign_language_translator.models.language_models.transformer_language_model.model import (
        TransformerLanguageModel,
    )


class LM_Dataset(torch.utils.data.Dataset):  # type: ignore
    def __init__(self, data: torch.Tensor):
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return (self.data[idx][..., :-1], self.data[idx][..., 1:])

    @staticmethod
    def prepare(
        file_path: str,
        text_to_token_ids: Callable[[str], List[int]],
        # padding_token_id: int,
        max_sequence_length: int = 32,
        encoding="utf-8",
        dtype=torch.int,
    ) -> List[torch.Tensor]:
        """Process a text file into list of 2d torch tensors of shape (n_examples, n_tokens).

        Args:
            file_path (str): where the input file is stored
            text_to_token_ids (Callable[[str], List[int]]): a function that can process a line from file and convert it into a list of token ids.
            max_sequence_length (int, optional): make n_grams of sequences longer than this of size max_sequence_length. Defaults to 32.
            encoding (str, optional): the encoding used in the text file. Defaults to "utf-8".
            dtype (_type_, optional): the type of returned torch tensors. check the range of values a type can contain and choose the smallest to save space. Defaults to torch.int.

        Returns:
            List[torch.Tensor]: _description_
        """

        with open(file_path, "r", encoding=encoding) as f:
            lines = f.read().splitlines()[:10000]

        def text_to_examples(text: str) -> List[List[int]]:
            ids = text_to_token_ids(text)
            # ids = ids + [padding_token_id] * max(0, max_sequence_length - len(ids) + 1)
            examples = (
                make_ngrams(ids, max_sequence_length + 1)
                if len(ids) > max_sequence_length + 1
                else [ids]
            )

            return examples  # type:ignore

        buckets = {}
        for examples in map(text_to_examples, tqdm(lines, leave=False)):
            for line in examples:
                if len(line) not in buckets:
                    buckets[len(line)] = []
                buckets[len(line)].append(line)

        data = [
            torch.Tensor(buckets[i]).type(dtype)  # torch.pad
            for i in range(2, max(buckets.keys()))  # type: ignore
            if i in buckets
        ]

        return data


class LM_Trainer:
    """class contains functions to train a language model built with pytorch.
    It is not designed to be generic rather it is specific to the TransformerLanguageModel.
    """

    def __init__(
        self,
        model: TransformerLanguageModel,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        epochs: int = 10,
        learning_rate: float = 1e-3,
        lr_lambda: Callable[
            [int, float, float], float
        ] = lambda epoch, base_lr, last_lr: base_lr,
        lr_update_step_count: int | None = None,
        optimizer="adamw",
        seed: int = 0,
        model_output_renderer: Callable[[TransformerLanguageModel], str] | None = None,
        epoch_unfreeze_map: Dict[int, List[str]] | None = None,
        class_weights: torch.Tensor | None = None,
        max_gradient_norm: float | None = None,
    ):
        self.model = model.to(device)
        self.device = device
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.optimizer = (
            torch.optim.AdamW(model.parameters(), lr=learning_rate)
            if optimizer == "adamw"
            else torch.optim.Adam(model.parameters(), lr=learning_rate)
        )
        self.scheduler = FullyLambdaLR(self.optimizer, lr_lambda=lr_lambda)

        self.seed = seed
        self.model_output_renderer = model_output_renderer

        self.lr_update_step_count = lr_update_step_count
        self.epoch_unfreeze_map = epoch_unfreeze_map
        self.class_weights = class_weights
        self.max_gradient_norm = max_gradient_norm

    def train(self, input_sequences, outputs) -> Dict[str, float]:
        """Training loop. calculates loss on non-padding tokens.
        Multiples class weights with the loss to scale it for imbalanced token distributions.
        Clips gradients with norm > max_gradient_norm to avoid exploding gradients.

        Args:
            input_sequences (_type_): batch of sequences
            outputs (_type_): batch of target sequences in which each position contains target token for the input sequence upto that position.

        Returns:
            Dict[str, float]: the metrics tracked e.g. loss
        """
        history = {}
        # inference
        self.model.train()
        self.optimizer.zero_grad()
        logits = self.model.forward(input_sequences)

        # reshape according to loss function's requirement
        B, T, C = logits.shape
        logits = logits.view(B * T, C)
        targets = outputs.reshape(B * T)

        # calculate loss
        not_padding_mask = targets != self.model.padding_token_id
        loss = torch.nn.functional.cross_entropy(
            logits[not_padding_mask], targets[not_padding_mask], reduction="none"
        )
        history["loss"] = loss.mean().item()

        # data
        if self.class_weights is not None:
            loss_multiplier = self.class_weights[targets[not_padding_mask]]
            loss_multiplier -= loss_multiplier.mean()
            loss_multiplier /= loss_multiplier.std()
            loss_multiplier -= loss_multiplier.min() - 1
            loss *= loss_multiplier
            history["scaled_loss"] = loss.mean().item()

        # back-propagation
        loss.mean().backward()

        if self.max_gradient_norm:
            torch.nn.utils.clip_grad_norm_(  # type:ignore
                self.model.parameters(), self.max_gradient_norm
            )

        self.optimizer.step()

        return history

    def run(
        self,
        train_batches: Iterable[Tuple[torch.Tensor, torch.Tensor]],
        validation_batches: Iterable[Tuple[torch.Tensor, torch.Tensor]],
        early_stop: bool = False,
        checkpoint_dir: str = "",
        checkpoint_step_count: int = 1000,
        model_output_step_count: int = 100,
        start_epoch_number: int = 0,
    ) -> Dict[str, List[float]]:
        """Run the training/validation loop and generate output & checkpoints.

        Returns:
            Dict[str, List[float]]: the tracked metrics
        """

        history = dict(train_loss=[], validation_loss=[])
        model_output = ""
        torch.manual_seed(self.seed)
        for epoch in range(start_epoch_number, start_epoch_number + self.epochs):
            # configure model
            set_layers_trainability_(
                self.model,
                layers_to_unfreeze=(self.epoch_unfreeze_map or {}).get(epoch, []),
            )

            # training loop
            start_time = time()
            train_losses = []
            for i, (input_sequences, outputs) in enumerate(train_batches):
                # move batch to right device
                input_sequences = input_sequences.to(self.device)
                outputs = outputs.to(self.device)

                # forward & back propagation
                _history = self.train(input_sequences, outputs)

                # verbose
                train_losses.append(_history["loss"])
                if i % model_output_step_count == 0:
                    with torch.no_grad():
                        self.model.eval()
                        model_output = (
                            f" | {self.model_output_renderer(self.model)}"
                            if self.model_output_renderer
                            else ""
                        )
                        self.model.train()
                remaining_seconds = (
                    (len(train_batches) - i - 1) * (time() - start_time) / (i + 1)  # type: ignore
                )
                print(
                    "\r"
                    f"epoch: {epoch} | "
                    f"train loss: {train_losses[-1]:.4f} | "
                    f"batches: {i+1}/{len(train_batches)} {(i+1)/len(train_batches):.1%} | "  # type: ignore
                    f"ETA: {remaining_seconds/(60 if remaining_seconds < 60**2 else 60 ** 2):.1f} {'min' if remaining_seconds < 60**2 else 'hr'} | "
                    f"LR: {self.scheduler.get_last_lr()[0]:.8f}" + model_output,
                    end="",
                )

                # save model
                if i % checkpoint_step_count == 0:
                    self.checkpoint(
                        checkpoint_dir,
                        losses=train_losses[-checkpoint_step_count:],
                        epoch=epoch,
                        steps_fraction=i / len(train_batches),  # type: ignore
                    )

                # Learning rate update
                if (not self.lr_update_step_count and i == len(train_batches) - 1) or (  # type: ignore
                    self.lr_update_step_count
                    and i % self.lr_update_step_count == self.lr_update_step_count - 1
                ):
                    self.scheduler.step()

            # mean_train_loss = np.mean(train_losses)
            mean_train_loss = self._average(train_losses)
            history["train_loss"].extend(train_losses)

            # validation loop
            start_time = time()
            val_losses = []
            with torch.no_grad():
                for i, (input_sequences, outputs) in enumerate(validation_batches):
                    # move batch to right device
                    input_sequences = input_sequences.to(self.device)
                    outputs = outputs.to(self.device)

                    # forward propagation
                    _history = self.validate(input_sequences, outputs)

                    # verbose
                    val_losses.append(_history["loss"])
                    print(
                        "\r"
                        f"epoch: {epoch} | "
                        f"train loss: {np.mean(train_losses[-1]):.4f} | "
                        f"val loss: {val_losses[-1]:.4f} | "
                        f"batches: {i+1}/{len(validation_batches)}= {(i+1)/len(validation_batches):.1%} | "  # type: ignore
                        f"ETA: {(len(validation_batches)-i-1)*(time()-start_time)/(i+1)/60:.1f} min",  # type: ignore
                        end="",
                    )
                mean_val_loss = np.mean(val_losses)
                history["validation_loss"].extend(val_losses)
                self.model.training_history["val_epoch"].append(epoch + 1)
                self.model.training_history["val_loss"].append(mean_val_loss)

            print(
                "\r"
                f"epoch: {epoch} | "
                f"train loss: {mean_train_loss:.4f}, "
                f"std: {np.std(train_losses):.4f} | "
                f"val loss: {mean_val_loss:.4f}, "
                f"std: {np.std(val_losses):.4f} | "
                f"LR: {self.scheduler.get_last_lr()[0]:.1e}"
            )
            if self.model_output_renderer:
                print("model output:", self.model_output_renderer(self.model))

            if early_stop:
                if mean_train_loss < mean_val_loss * 0.95:
                    print("Early stopping")
                    break

        return history

    def validate(
        self, input_sequences: torch.Tensor, outputs: torch.Tensor
    ) -> Dict[str, float]:
        """the validation loop. infers the model on validation data without gradients or back propagation and calculates metrics.

        Args:
            input_sequences (torch.Tensor): batch of input tokens
            outputs (torch.Tensor): batch of target sequences

        Returns:
            Dict[str, float]: the tracked metrics
        """
        history = {}
        with torch.no_grad():
            # inference
            self.model.eval()
            logits = self.model.forward(input_sequences)

            # reshape according to loss function's requirement
            B, T, C = logits.shape
            logits = logits.view(B * T, C)
            targets = outputs.reshape(B * T)

            # calculate loss
            not_padding_mask = targets != self.model.padding_token_id
            loss = torch.nn.functional.cross_entropy(
                logits[not_padding_mask], targets[not_padding_mask], reduction="none"
            )
            history["loss"] = loss.mean().item()

            # handle data imbalance
            if self.class_weights is not None:
                loss_multiplier = self.class_weights[targets[not_padding_mask]]
                loss_multiplier -= loss_multiplier.mean()
                loss_multiplier /= loss_multiplier.std()
                loss_multiplier -= loss_multiplier.min() - 1
                loss *= loss_multiplier
                # loss *= self.class_weights[targets[not_padding_mask]]
                history["scaled_loss"] = loss.mean().item()

        return history

    def checkpoint(self, checkpoint_dir: str, losses, epoch, steps_fraction) -> None:
        """save metrics in model and save model to disk."""

        self.model.training_history["train_epoch"].append(epoch + steps_fraction)
        avg_loss = self._average(losses)
        self.model.training_history["train_loss"].append(avg_loss)

        if os.path.isdir(checkpoint_dir):
            previous_checkpoint_paths = glob(
                os.path.join(
                    checkpoint_dir,
                    f"checkpoint_tlm_{self.model.n_parameters/1e6:.1f}M_*loss.pt",
                )
            )
            previous_best_loss = min(
                float(p.split("M_")[-1][:-7])
                for p in previous_checkpoint_paths + ["0M_inf1234567"]
            )
            if avg_loss > previous_best_loss:
                return
            for p in previous_checkpoint_paths:
                os.remove(p)
            self.model.save(
                os.path.join(
                    checkpoint_dir,
                    f"checkpoint_tlm_{self.model.n_parameters/1e6:.1f}M_{avg_loss:.4f}loss.pt",
                )
            )
        else:
            print("checkpoint failed, dir not found")

    def _average(self, values) -> float:
        return np.average(values, weights=np.geomspace(1, 20, len(values)))


__all__ = [
    "LM_Dataset",
    "LM_Trainer",
]


if __name__ == "__main__":
    # TODO: write training script (for now, see notebooks repo)
    # https://github.com/sign-language-translator/notebooks/blob/main/model_training/transformer_lm_training.ipynb
    pass



from typing import Any, Iterable, List, Set, Tuple, Union

from sign_language_translator.config.enums import TextLanguages
from sign_language_translator.languages.text.text_language import TextLanguage
from sign_language_translator.languages.vocab import Vocab


class English(TextLanguage):
    """NLP class for English text. Extends `slt.languages.text.TextLanguage` class.

    English is originally a West Germanic language and potentially an international language in the 21st century. English uses the Latin script, which consists of 26 letters and is written from left to right.
    See unicode details at: https://unicode.org/charts/PDF/U0000.pdf
    """

    @staticmethod
    def name() -> str:
        raise NotImplementedError
        # return TextLanguages.ENGLISH.value

    @classmethod
    def token_regex(cls) -> str:
        raise NotImplementedError

    @classmethod
    def allowed_characters(cls) -> Set[str]:
        raise NotImplementedError

    def __init__(self) -> None:
        raise NotImplementedError
        # self.vocab = Vocab(
        #     language=f"^{self.name()}$",  # r"^en$"
        #     country=r"[a-z]+",
        #     organization=r"[a-z]+",
        #     part_number=r"[0-9]+",
        #     data_root_dir=Assets.ROOT_DIR,
        #     arg_is_regex=True,
        # )

    def preprocess(self, text: str) -> str:
        raise NotImplementedError

    def tokenize(self, text: str) -> List[str]:
        raise NotImplementedError

    def sentence_tokenize(self, text: str) -> List[str]:
        raise NotImplementedError

    def detokenize(self, tokens: Iterable[str]) -> str:
        raise NotImplementedError

    def tag(self, tokens: Union[str, Iterable[str]]) -> List[Tuple[str, Any]]:
        raise NotImplementedError

    def get_tags(self, tokens: Union[str, Iterable[str]]) -> List[Any]:
        raise NotImplementedError

    def get_word_senses(self, tokens: Union[str, Iterable[str]]) -> List[List[str]]:
        raise NotImplementedError



"""Centralized String Constants for Sign Language Translator Package

This module defines enumerations for various string constants used throughout the package.

Enumerations:
    Countries (Enum): Enumerates supported countries with their short codes.
    Organizations (Enum): Enumerates supported organizations with their short codes.
    SignCollections (Enum): Enumerates sign collections with their names.
    TextLanguages (Enum): Enumerates supported text languages with their short codes.
    SignLanguages (Enum): Enumerates supported sign languages with their names.
    SignFormats (Enum): Enumerates supported video feature models.
    ModelCodes (Enum): Enumerates model codes for different models.

Functions:
    normalize_short_code (function): Normalizes a provided short code to a recognized standard form.
"""

from __future__ import annotations

from enum import Enum

from sign_language_translator.utils import (
    PrintableEnumMeta,
    search_in_values_to_retrieve_key,
)

__all__ = [
    "Countries",
    "Organizations",
    "SignCollections",
    "TextLanguages",
    "SignLanguages",
    "SignFormats",
    "ModelCodes",
    "normalize_short_code",
]


class Countries(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of countries with their corresponding short codes.

    Attributes:
        - PAKISTAN (str): Short code for Pakistan.
        ...
    """

    PAKISTAN = "pk"
    # INDIA = "in"
    # USA = "us"


class Organizations(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of organizations with their corresponding short codes.

    Attributes:
        - HFAD (str): Short code for HFAD (e.g., an organization for deaf in Pakistan).
        ...
    """

    HFAD = "hfad"
    # NISE = "nise"
    # FESF = "fesf"


class SignCollections(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of sign collections with their corresponding short codes.

    Attributes:
        - PK_HFAD_1 (str): Short code for the first sign collection in Pakistan made in HFAD.
        - WORDLESS (str): Short code for a sign collection with videos containing no signs.
        ...
    """

    PK_HFAD_1 = f"{Countries.PAKISTAN}-{Organizations.HFAD}-1"
    WORDLESS = "wordless"


class TextLanguages(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of text languages with their corresponding short codes.

    Attributes:
        - URDU (str): Short code for the Urdu language.
        - HINDI (str): Short code for the Hindi language.
        ...
    """

    URDU = "ur"
    # ENGLISH = "en"
    HINDI = "hi"


class SignLanguages(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of sign languages with their corresponding short codes.

    Attributes:
        - PAKISTAN_SIGN_LANGUAGE (str): Short code for the Pakistan Sign Language.
        ...
    """

    PAKISTAN_SIGN_LANGUAGE = "pakistan-sign-language"


class SignFormats(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of available sign formats with their corresponding short codes.
    For example, sign language can be a sequence of frames (video) or a sequence of pose vectors (landmarks) etc.

    Attributes:
        - VIDEO (str): Short code for raw video.
        - MEDIAPIPE_LANDMARKS (str): Short code for MediaPipe Pose & Hand landmarks.
        ...
    """

    # raw video
    VIDEO = "video"
    """sequence of numpy frames (num_frames, height, width, num_channels)"""

    # Body Landmarks
    MEDIAPIPE_LANDMARKS = "mediapipe-landmarks"
    """3d world coordinates (x,y,z,[visibility, presence]) and/or 2d image coordinates (x,y,depth,[visibility, presence])"""

    # Body Mesh Grid

    # Image Segmentation

    # Motion Vectors


class ModelCodes(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of model codes with their corresponding short codes.

    Attributes:
        - CONCATENATIVE_SYNTHESIS (str): Short code for the our rule-based text-to-sign translation model.

        - NGRAM_LM_UNIGRAM_NAMES (str): Short code for ngram model trained with window size 1 on en/ur person names data.
        - NGRAM_LM_BIGRAM_NAMES (str): Short code for ngram model trained with window size 2 on en/ur person names data.
        - NGRAM_LM_TRIGRAM_NAMES (str): Short code for ngram model trained with window size 3 on en/ur person names data.
        - MIXER_LM_NGRAM_URDU (str): Short code for a mix of ngram models trained on urdu words of window size 1 to 6.
        - TRANSFORMER_LM_UR_SUPPORTED (str): Short code for a transformer-based language model trained on ur supported tokens.

        - MEDIAPIPE_POSE_V2_HAND_V1 (str): Short code for a video embedding model which uses MediaPipe pose_landmarker_heavy & hand_landmarker.
        - MEDIAPIPE_POSE_V1_HAND_V1 (str):
        - MEDIAPIPE_POSE_V0_HAND_V1 (str):
        ...
    """

    # text-to-sign translation
    CONCATENATIVE_SYNTHESIS = "concatenative-synthesis"
    """Short code for the core rule-based text to sign translation model that joins video clips for each word in a sentence."""
    # LANDMARK_GAN = "landmark-gan"

    # sign-to-text translation
    # GESTURE = "gesture"

    # language-models
    NGRAM_LM_UNIGRAM_NAMES = "names-stat-lm-w1.json"
    NGRAM_LM_BIGRAM_NAMES = "names-stat-lm-w2.json"
    NGRAM_LM_TRIGRAM_NAMES = "names-stat-lm-w3.json"
    MIXER_LM_NGRAM_URDU = "ur-supported-token-unambiguous-mixed-ngram-w1-w6-lm.pkl"
    TRANSFORMER_LM_UR_SUPPORTED = "tlm_14.0M.pt"

    # video-embedding-models
    MEDIAPIPE_POSE_V2_HAND_V1 = "mediapipe-pose-2-hand-1"
    MEDIAPIPE_POSE_V1_HAND_V1 = "mediapipe-pose-1-hand-1"
    MEDIAPIPE_POSE_V0_HAND_V1 = "mediapipe-pose-0-hand-1"

    # text-embedding-models

class ModelCodeGroups(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration class for grouping supported model codes, making it easier to filter various models.

    Attributes:
        ALL_LANGUAGE_MODELS (set): Set of model codes for all language models.
        ALL_VIDEO_FEATURE_MODELS (set): Set of model codes for all video embedding models.

        ALL_NGRAM_LANGUAGE_MODELS (set): Set of model codes for all n-gram language models.
        ALL_TRANSFORMER_LANGUAGE_MODELS (set): Set of model codes for all transformer-based language models.
        ALL_MIXER_LANGUAGE_MODELS (set): Set of model codes for all mixer-based language models.

        ALL_MEDIAPIPE_EMBEDDING_MODELS (set): Set of model codes for all MediaPipe-based video embedding models.
        ALL_VIDEO_EMBEDDING_MODELS (set): Set of model codes for all video embedding models.
    """

    # language models
    ALL_NGRAM_LANGUAGE_MODELS = {
        ModelCodes.NGRAM_LM_UNIGRAM_NAMES.value,
        ModelCodes.NGRAM_LM_BIGRAM_NAMES.value,
        ModelCodes.NGRAM_LM_TRIGRAM_NAMES.value,
    }
    ALL_TRANSFORMER_LANGUAGE_MODELS = {
        ModelCodes.TRANSFORMER_LM_UR_SUPPORTED.value,
    }
    ALL_MIXER_LANGUAGE_MODELS = {
        ModelCodes.MIXER_LM_NGRAM_URDU.value,
    }
    ALL_LANGUAGE_MODELS = (
        ALL_NGRAM_LANGUAGE_MODELS
        | ALL_TRANSFORMER_LANGUAGE_MODELS  # type: ignore
        | ALL_MIXER_LANGUAGE_MODELS
    )

    # video embedding models
    ALL_MEDIAPIPE_EMBEDDING_MODELS = {
        ModelCodes.MEDIAPIPE_POSE_V2_HAND_V1.value,
        ModelCodes.MEDIAPIPE_POSE_V1_HAND_V1.value,
        ModelCodes.MEDIAPIPE_POSE_V0_HAND_V1.value,
    }
    ALL_VIDEO_EMBEDDING_MODELS = ALL_MEDIAPIPE_EMBEDDING_MODELS

# TODO: move the mapping list outside of the function. maybe convert it to a class variable.
def normalize_short_code(short_code: str | Enum) -> str:
    """
    Normalize the provided short code to a standard form that is recognized package wide.

    Args:
        short_code (str): The short code to be normalized.

    Returns:
        str: The normalized short code.
    """

    if isinstance(short_code, Enum):
        short_code = str(short_code.value)

    normalized_to_codes = {
        ModelCodes.CONCATENATIVE_SYNTHESIS.value: {
            "rule-based",
            "concatenative",
            "concatenativesynthesis",
            "concatenative-synthesis",
            "concatenative_synthesis",
            "concat-synth",
        },
        TextLanguages.URDU.value: {
            "urdu",
            "urd",
            "ur",
        },
        TextLanguages.HINDI.value: {
            "hindi",
            "hin",
            "hi",
        },
        SignLanguages.PAKISTAN_SIGN_LANGUAGE.value: {
            "psl",
            "pk-sl",
            "pakistan-sign-language",
            "pakistan-sl",
            "pakistansignlanguage",
            "pakistan_sign_language",
        },
        SignFormats.VIDEO.value: {
            "vid",
            "videos",
            "vids",
        },
        ModelCodes.NGRAM_LM_UNIGRAM_NAMES.value: {
            "unigram-names",
        },
        ModelCodes.NGRAM_LM_BIGRAM_NAMES.value: {
            "bigram-names",
        },
        ModelCodes.NGRAM_LM_TRIGRAM_NAMES.value: {
            "trigram-names",
        },
        ModelCodes.MIXER_LM_NGRAM_URDU.value: {
            "urdu-mixed-ngram",
            "urdu-supported-ngram",
            "ur-mixed-ngram",
            "ur-supported-ngram",
        },
        ModelCodes.TRANSFORMER_LM_UR_SUPPORTED.value: {
            "ur-supported-gpt",
            "urdu-supported-gpt",
        },
        ModelCodes.MEDIAPIPE_POSE_V2_HAND_V1.value: {
            "mediapipe",
        },
    }
    normalized_to_codes = {
        k: v.union(
            {
                k,
                k.replace("-", "_"),
                k.replace("_", "-"),
                *map(lambda x: x.replace("-", "_"), v),
            }
        )
        for k, v in normalized_to_codes.items()
    }

    normalized = search_in_values_to_retrieve_key(short_code, normalized_to_codes)

    return normalized or short_code



"""Module for managing assets required by the sign language translator package,
such as datasets and models.
"""

import json
import os
import re
from datetime import datetime
from os.path import abspath, dirname, exists, isdir, isfile, join, sep
from typing import Dict, List, Tuple

from tqdm.auto import tqdm

from sign_language_translator.config.settings import Settings
from sign_language_translator.config.utils import read_urls
from sign_language_translator.utils import (
    Archive,
    ProgressStatusCallback,
    download,
    is_regex,
)


class Assets:
    """
    Static class for managing assets related to the sign language translator.
    It wraps around utility functions to automatically handle downloading, extracting, deleting and loading assets.

    Attributes:
        ROOT_DIR (str): The root directory path where the sign language datasets & models are stored.
        FILE_TO_URL (Dict[str, str]): A dictionary mapping asset filenames to their corresponding URLs.
        asset_regex_to_urls_file (List[Tuple[str, str]]): A list of tuples mapping regular expressions matching asset names to URLs files. The UR file must be a JSON containing a key "file_to_url" which maps to an object mapping relative paths of assets to their URLs.

    Methods:
        set_root_dir(path: str) -> None:
            Set the SLT resources directory path.

        get_ids(filename_or_regex: str) -> List[str]:
            Get the relative paths of assets matching the given filename_or_regex.

        download(filename_or_regex: str, overwrite=False, progress_bar: bool | None = None, timeout: float = 20.0, leave=True, chunk_size=65536) -> bool:
            Download assets matching the given filename regex and save them to the appropriate file paths inside the assets root directory.

        extract(filename_or_regex: str, archive_name_or_regex: str | None = None, overwrite=False, download_archive=True) -> List[str]:
            extract the files matching the argument from an archived dataset into the appropriate location.

        fetch(filename_or_regex: str, overwrite=False, archive_name_or_regex: str=None, download_archive=False) -> List[str]:
            extract the required files if an archive is pre-downloaded otherwise download the files if direct URL is available otherwise download the archive and extract.

        delete(filename_or_regex: str) -> None:
            remove the matching assets from storage and its records from the checksum file.

    Example:
    .. code-block:: python

        import sign_language_translator as slt
        # slt.Assets.set_root_dir("./custom-centralized-assets")

        # Archived datasets
        ids = slt.Assets.get_ids(r"datasets/.*\\.zip")
        paths = slt.Assets.download("datasets/pk-hfad-1_landmarks-mediapipe-pose-2-hand-1-csv.zip")
        files = slt.Assets.extract(paths[0], r".*\\.csv")

        # all dictionary videos for numbers
        urls = slt.Assets.get_url(r"videos/[a-z-]+_\\d+\\.mp4")

        # Automatically extract dictionary videos from an archive if its available else download from direct URL
        # try not to use vague regex anywhere and its best if you manually download needed dataset archive
        paths = slt.Assets.fetch(r"videos/pk-hfad-1_\\d+\\.mp4")

        # download a model
        paths = slt.Assets.download(r".*/names-stat-lm-w1\\.json")

        # add your own dataset
        slt.Assets.FILE_TO_URL.update({"relative/path/to/asset.mp4": "https://...",})
    """

    # =============== #
    #    CONFIGURE    #
    # =============== #

    ROOT_DIR: str = join(dirname(dirname(abspath(__file__))), "assets")
    """The root directory path where the sign language datasets & models are stored. Defaults is 'install_directory/assets'."""

    primary_urls_file: str = "urls.json"
    """The name of the first URLs file that is loaded by default and must contain links to other url files.
    Defaults is 'urls.json'. Note: all url filenames must end with 'urls.json'."""

    urls_file_dir: str = dirname(abspath(__file__))
    """The directory path where the URLs files are stored. Defaults is 'install_directory/sign_language_translator/config'."""

    FILE_TO_URL: Dict[str, str] = read_urls(join(urls_file_dir, primary_urls_file))
    """A dictionary mapping asset filenames to their corresponding URLs."""

    asset_regex_to_urls_file: List[Tuple[str, str]] = [
        # regex_for_assets, urls_file
        (r"^videos/pk-.*mp4$", "pk-dictionary-urls.json"),
        (r".*zip$", "archive-urls.json"),
        (r".*", "extra-urls.json"),
    ]
    """A list of tuples that map regular expressions matching asset names to URLs files containing URLs to that asset group.
    The regex are tried sequentially so make sure list order is right."""

    _checksum_filename = "checksum.json"
    _loaded_url_files = set()

    @classmethod
    def set_root_dir(cls, path: str) -> None:
        """Set the SLT resources directory path.
        Helpful when using custom datasets, or when data must be stored outside install directory at a centralized location.
        (e.g hosted on cloud and mounted on disk).
        By default, resources are downloaded to the 'install_directory/assets'.

        Args:
            path (str): The path to the assets, datasets or models directory.

        Raises:
            ValueError: If the provided path is not a directory.
        """

        path = abspath(path)
        if exists(path) and not isdir(path):
            raise ValueError(f"The provided path is not a directory. Path: {path}")
        os.makedirs(path, exist_ok=True)

        cls.ROOT_DIR = path

    # ============= #
    #    Getters    #
    # ============= #

    @classmethod
    def get_ids(cls, filename_or_regex: str) -> List[str]:
        """
        Filters down the loaded Assets.FILE_TO_URLS dictionary and
        returns the list of asset IDs (relative paths) that match the given filename_or_regex.
        If no asset_id match the argument, appropriate urls file is loaded from the
        Assets.asset_regex_to_urls_file list and the function is called recursively.

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.

        Returns:
            List[str]: List of matching asset IDs.
        """

        # exact match
        if filename_or_regex in cls.FILE_TO_URL:
            return [filename_or_regex]

        # regex match #* (but only with currently loaded url files)
        if is_regex(filename_or_regex):
            regex = re.compile(f"^{filename_or_regex}$")
            if ids := [file for file in cls.FILE_TO_URL if regex.match(file)]:
                return ids

        # load more urls
        for regex, urls_file in cls.asset_regex_to_urls_file:
            if urls_file in cls._loaded_url_files:
                continue

            if regex == filename_or_regex or re.match(regex, filename_or_regex):
                cls.load_urls(urls_file)

                # recursive call
                return cls.get_ids(filename_or_regex)

        # not found
        return []

    @classmethod
    def get_url(cls, filename_or_regex: str) -> List[str]:
        """
        Filters down the loaded Assets.FILE_TO_URLS dictionary and
        returns the list of URLs corresponding to the asset_ids matching given filename_or_regex.

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.

        Returns:
            List[str]: List of matching URLs.
        """

        return [cls.FILE_TO_URL[id] for id in cls.get_ids(filename_or_regex)]

    @classmethod
    def get_path(cls, filename_or_regex: str) -> List[str]:
        """
        Filters down the loaded Assets.FILE_TO_URLS dictionary and
        returns the list of file paths corresponding to the asset_ids matching given filename_or_regex.

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.

        Returns:
            List[str]: List of matching file paths.
        """

        return [cls._abs_path(asset_id) for asset_id in cls.get_ids(filename_or_regex)]

    # ============== #
    #    Fetching    #
    # ============== #

    @classmethod
    def extract(
        cls,
        filename_or_regex: str,
        archive_name_or_regex: str | None = None,
        overwrite=False,
        progress_bar=True,
        leave=True,
        download_archive=True,
    ) -> List[str]:
        """
        Extracts assets matching the given filename_or_regex from the archived datasets.
        The target location is inferred from the archive & asset name and resides inside assets root directory.
        If the archive name is not provided, it will be inferred from the filename_or_regex, so avoid using vague regex and follow the filename structure.
        If the archive is not downloaded, it will be downloaded first.

        Note:
            Please use the `slt.utils.Archive.extract()` function directly if you want deterministic behavior and avoid false guesses.

        Args:
            filename_or_regex (str): The filename or regex pattern to match the archive contents.
            archive_name_or_regex (str, optional): The name or regex pattern of the archive(s) that contains the assets. If None, the function tries guess it from the content name. Defaults to None.
            overwrite (bool, optional): Flag indicating whether to overwrite existing assets. Defaults to False.
            progress_bar (bool, optional): Flag indicating whether to display a progress bar during extraction. Defaults to True.
            leave (bool, optional): Flag indicating whether to leave the progress bar displayed after extraction. Defaults to True.
            download_archive (bool, optional): Flag indicating whether to download the archive if it is not already downloaded. Defaults to True.

        Returns:
            List[str]: A list of paths to the extracted assets.
        """
        cls.delete_out_of_date_assets()

        archive_regex = archive_name_or_regex or cls._infer_archive_name(
            filename_or_regex
        )
        if download_archive:
            cls.download(
                archive_regex, overwrite=False, progress_bar=progress_bar, leave=leave
            )

        content_name = filename_or_regex.split("/")[-1]
        extracted_assets = []
        for archive_id in cls.get_ids(archive_regex):
            asset_dir = archive_id.split("/")[-1].split("_")[-1].split("-")[0]
            extracted_assets += Archive.extract(
                archive_path=cls._abs_path(archive_id),
                regex=content_name,
                output_dir=cls._abs_path(asset_dir),
                overwrite=overwrite,
                progress_bar=progress_bar,
                leave=leave,
            )

        # TODO: checksum

        return extracted_assets

    @classmethod
    def download(
        cls,
        filename_or_regex: str,
        overwrite=False,
        timeout: float = 20.0,
        chunk_size=2**18,
        progress_bar: bool | None = None,
        leave=True,
    ) -> List[str]:
        """
        Downloads package assets matching the given filename regex and saves them to the appropriate file paths.

        Args:
            filename_or_regex (str): Relative path or Regular expression to match the desired asset names.
            overwrite (bool, optional): If False, skips downloading if the resource file already exists. Defaults to False.
            timeout (float, optional): The maximum number of seconds to wait for a server response. Defaults to 20.0.
            chunk_size (int, optional): The number of bytes to fetch in each step. Defaults to 256*1024.
            progress_bar (bool, optional): If True, displays a progress bar during the download. If None, uses the value in slt.Settings.SHOW_DOWNLOAD_PROGRESS. Defaults to None.
            leave (bool, optional): Wether to leave the progress bar behind after the download. Defaults to True.

        Returns:
            List[str]: List of paths to matching files that were downloaded or existed already.
        """
        cls.delete_out_of_date_assets()
        existing_paths = []

        # Select assets to download
        id_path_url = []
        for asset_id in cls.get_ids(filename_or_regex):
            path = cls._abs_path(asset_id)
            if exists(path) and not overwrite:
                existing_paths.append(path)
                continue
            id_path_url.append((asset_id, path, cls.FILE_TO_URL[asset_id]))

        # Configure progress bar
        callback = None
        if progress_bar is None:
            progress_bar = Settings.SHOW_DOWNLOAD_PROGRESS
            leave = False
        if progress_bar and len(id_path_url) > 1:
            id_path_url = tqdm(id_path_url, leave=leave)
            callback = ProgressStatusCallback(id_path_url)

        # Download assets
        for asset_id, file_path, url in id_path_url:
            # progress bar
            if isinstance(id_path_url, tqdm):
                id_path_url.set_description(f"Downloading {asset_id}")

            # Make sure that the file/directory exists
            os.makedirs(dirname(file_path), exist_ok=True)

            # Download the file from the URL
            success_in_download = download(
                file_path,
                url,
                progress_bar=(progress_bar and len(id_path_url) == 1),
                timeout=timeout,
                overwrite=overwrite,
                chunk_size=chunk_size,
                status_callback=callback,
            )

            # update checksum file with date, url, filename, n_bytes
            if success_in_download:
                cls._update_checksum(asset_id, {"url": url})

        return [path for _, path, _ in id_path_url] + existing_paths

    @classmethod
    def fetch(
        cls,
        filename_or_regex: str,
        overwrite=False,
        archive_name_or_regex: str | None = None,
        download_archive=False,
        progress_bar=True,
        leave=True,
    ) -> List[str]:
        """Tries to extract the asset from an archive if it is available, otherwise downloads the asset from its direct URL if available.
        If the archive name is not provided, it will be inferred from the filename_or_regex.

        Note:
            Only use for dictionary videos because only their direct URLs & archives both are available.
            Follow the filename structure to avoid false guesses (i.e. use all - and _ in the right places and insert wildcards as required).

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.
            overwrite (bool, optional): Flag indicating whether to overwrite existing assets. Defaults to False.
            archive_name_or_regex (str, optional): The name or regex pattern of the archive(s) that contains the assets. If None, the function tries guess it from the content name. Defaults to None.
            download_archive (bool, optional): Flag indicating whether to download the archive if it is not already downloaded. Defaults to False.
            progress_bar (bool, optional): Flag indicating whether to show the progress bars or not. Defaults to True.
            leave (bool, optional): Flag indicating whether to leave the progress_bar behind after completion or not. Defaults to True.

        Returns:
            List[str]: List of paths to files that were downloaded or extracted.

        Example:
        .. code-block:: python

            import sign_language_translator as slt

            # extract/download all pakistan-sign-language dictionary videos for numbers
            paths = slt.Assets.fetch(r"videos/pk-[a-z]+-[0-9]+_\\d+\\.mp4")
        """

        extracted_assets = cls.extract(
            filename_or_regex,
            archive_name_or_regex=archive_name_or_regex,
            overwrite=overwrite,
            download_archive=download_archive,
            progress_bar=progress_bar,
            leave=leave,
        )
        downloaded_assets = cls.download(
            filename_or_regex,
            overwrite=False,
            progress_bar=progress_bar,
            leave=leave,
        )

        return list(set(extracted_assets + downloaded_assets))

    # ============ #
    #    Delete    #
    # ============ #

    @classmethod
    def delete_out_of_date_assets(cls) -> None:
        """
        Delete asset if it is out of date. Currently determined by comparing the download URL
        of the asset in checksum file with the loaded URL.
        Does not delete files not present in checksum.
        """

        checksum = cls._read_checksum()

        for asset_id, info in checksum.copy().items():
            path = cls._abs_path(asset_id)
            if not exists(path):
                checksum.pop(asset_id, None)
                continue
            if (
                "url" in info
                and asset_id in cls.FILE_TO_URL  # is loaded
                and info["url"] != cls.FILE_TO_URL[asset_id]
            ):
                os.remove(path)
                checksum.pop(asset_id, None)

            # TODO: check info["archive_id"] and delete archive & all extracted_contents

        cls._write_checksum(checksum)

    @classmethod
    def delete(cls, filename_or_regex: str):
        checksum = cls._read_checksum()
        matching_ids = cls.get_ids(filename_or_regex)
        for asset_id in matching_ids:
            if exists(file_path := cls._abs_path(asset_id)):
                os.remove(file_path)
                checksum.pop(asset_id, None)

        cls._write_checksum(checksum)

    # ================= #
    #    URL Loading    #
    # ================= #

    @classmethod
    def reload(cls) -> None:
        """
        Clear the cache and read the URL files again.
        """

        cls.FILE_TO_URL = read_urls(join(cls.urls_file_dir, cls.primary_urls_file))
        cls._loaded_url_files = set()

    @classmethod
    def load_urls(cls, filename: str) -> None:
        """
        Load URLs from the specified file into the Assets.FILE_TO_URL dictionary.

        Args:
            filename (str): The name of the URLs file to load.
        """
        cls.download(filename)
        for file_id in cls.get_ids(filename):
            cls.FILE_TO_URL.update(read_urls(cls._abs_path(file_id)))
            cls._loaded_url_files.add(file_id)

    @classmethod
    def load_all_urls(cls) -> None:
        """
        Load all URL files into the Assets.FILE_TO_URL dictionary.
        """
        for _, urls_file in cls.asset_regex_to_urls_file:
            cls.load_urls(urls_file)
        cls.load_urls(r"([a-z-]+[-_])?urls.json")

    # ============= #
    #    helpers    #
    # ============= #

    @classmethod
    def _abs_path(cls, asset_id: str) -> str:
        return join(
            (cls.urls_file_dir if asset_id.endswith("urls.json") else cls.ROOT_DIR),
            asset_id.replace("/", sep),
        )

    @classmethod
    def _read_checksum(cls) -> Dict[str, Dict[str, str]]:
        os.makedirs(cls.ROOT_DIR, exist_ok=True)
        checksum_file_path = join(cls.ROOT_DIR, cls._checksum_filename)
        if isfile(checksum_file_path):
            with open(checksum_file_path, "r", encoding="utf-8") as f:
                checksum: Dict[str, Dict[str, str]] = json.load(f)
        else:
            checksum = {}
            cls._write_checksum(checksum)

        return checksum

    @classmethod
    def _write_checksum(cls, checksum: Dict[str, Dict[str, str]]) -> None:
        checksum_file_path = join(cls.ROOT_DIR, cls._checksum_filename)
        with open(checksum_file_path, "w", encoding="utf-8") as f:
            json.dump(checksum, f, indent=2, ensure_ascii=False, sort_keys=True)

    @classmethod
    def _update_checksum(cls, asset_id: str, info: Dict):
        checksum = cls._read_checksum()
        if asset_id not in checksum:
            checksum[asset_id] = {}
        checksum[asset_id].update(
            {"date": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"), **info}
        )
        cls._write_checksum(checksum)

    @classmethod
    def _infer_archive_name(cls, filename_or_regex: str) -> str:
        """
        Infers the archive name/regex that should contain the given asset based on the provided filename_or_regex argument.

        Args:
            filename_or_regex (str): The asset filename or regex from which its containing archive name must be inferred.

        Returns:
            str: A regex pattern that matches the archive name which should contain the given asset.
        """

        if filename_or_regex.endswith("zip"):
            return filename_or_regex

        base = filename_or_regex.split("/")[-1]
        collection = base.split("_", 1)[0] if "_" in base else ".*"
        extension = ext if (ext := base[-3:]) in ("csv", "npz", "npy", "mp4") else ".*"

        if "video" in filename_or_regex or extension == "mp4":
            category = "videos"
            model = "?"
            # todo: remove this hack
            # todo: make it work for dictionary & replications
            if extension == ".*":
                extension = "mp4"
        elif "landmarks" in filename_or_regex:
            category = "landmarks"
            model = ".*"
        else:
            model = ".^"
            category = ".^"

        return f"datasets/{collection}_{category}-{model}-{extension}\\.zip"



"""
This module defines a Transformer-based Language Model that can be used for text generation and language modeling.

Class:
    TransformerLanguageModel: A class that implements a Transformer-based Language Model.
"""

from __future__ import annotations

from copy import deepcopy
from os.path import exists
from typing import Any, Dict, Iterable, List, Tuple

import numpy as np
import torch

from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.models.language_models.transformer_language_model.layers import (
    DecoderBlock,
)
from sign_language_translator.models.utils import top_p_top_k_indexes
from sign_language_translator.utils import sample_one_index


class TransformerLanguageModel(LanguageModel, torch.nn.Module):
    """Transformer-based language model for text generation.

    This class implements a Transformer-based language model for text generation tasks.
    It takes in a sequence of token IDs and generates the next token in the sequence.
    The model consists of two embedding layers, multiple decoder blocks, and a language modeling head.

    Attributes:
        - token_embedding (torch.nn.Embedding): The embedding layer for token IDs.
        - position_embedding (torch.nn.Embedding): The embedding layer for positional IDs.
        - decoder_blocks (torch.nn.Sequential): The sequence of decoder blocks.
        - final_layer_norm (torch.nn.LayerNorm): The layer normalization for the final output.
        - language_modeling_head (torch.nn.Linear): The linear layer for language modeling.
        - n_parameters (int): The total number of parameters in the model.
        - device (str): The device to run the model on.
        - training_history (Dict[str, Any]): The training history of the model such as loss and other metrics.

    Methods:
        - forward(token_ids: torch.Tensor) -> torch.Tensor: Performs a forward pass through the model.
        - next(self, context: Iterable) -> Tuple[Any, float]: generates the next token and its probability.
        - next_all(self, context: Iterable) -> Tuple[List[Any], List[float]]: returns all next tokens and their probabilities.
        - load(model_path: str) -> TransformerLanguageModel: (static_method) Deserializes the model from a pt file.
        - save(self, model_path: str, device: str | Torch.device): Serializes the model to a pt file.
        - get_model_state() -> Dict[str, Any]: Returns the model state consisting of constructor arguments and pytorch state_dict.
        - tokens_to_ids(tokens: Iterable[str]) -> List[int]: Converts tokens to IDs.
        - ids_to_tokens(ids: Iterable[int] | torch.Tensor) -> List[str]: Converts IDs to tokens.
    """

    def __init__(
        self,
        token_to_id: Dict[str, int],
        vocab_size: int,
        unknown_token="<unk>",
        padding_token="<pad>",
        start_of_sequence_token="<sos>",
        window_size: int = 64,
        embed_size: int = 768,
        hidden_size: int = 768 * 4,
        n_heads: int = 6,
        n_blocks: int = 6,
        dropout: float = 0.25,
        activation="gelu",
        device="cuda" if torch.cuda.is_available() else "cpu",
        sampling_temperature: float = 1.0,
        top_k: int | None = None,
        top_p: float | None = 0.9,
        name: str | None = None,
        pretrained_token_embeddings: torch.Tensor | None = None,
        randomly_shift_position_embedding_during_training: bool = False,
    ) -> None:
        """Transformer-based language model for text generation.

        Args:
            - token_to_id (Dict[str, int]): A dictionary mapping token strings to their corresponding IDs.
            - vocab_size (int): The size of the vocabulary. Determine model's dimensionality.
            - unknown_token (str): The token representing unknown words.
            - padding_token (str): The token representing padding.
            - start_of_sequence_token (str): The token representing the start of a sequence.
            - window_size (int): The maximum context length of input for positional embeddings.
            - embed_size (int): The size of the token embeddings.
            - hidden_size (int): The size of the hidden layers in the decoder blocks.
            - n_heads (int): The number of self attention heads in each decoder block.
            - n_blocks (int): The number of decoder blocks.
            - dropout (float): The dropout rate for regularization.
            - activation (str): The activation function used in the decoder blocks.
            - device (str): The device to run the model on (default is 'cuda' if available, else 'cpu').
            - sampling_temperature (float): The temperature for sampling from the output distribution. High means more randomness, low is more probable. Defaults to 1.0.
            - top_k (int | None): The number of top-k candidates to consider during sampling.
            - top_p (float | None): The threshold for top cumulative probability during sampling.
            - name (str | None): The name of the model.
            - pretrained_token_embeddings (torch.Tensor | None): Pretrained token embeddings.
            - randomly_shift_position_embedding_during_training (bool): Flag to shift the position ids by a random amount during training. Defaults to False.
        """

        LanguageModel.__init__(self, unknown_token, name)
        torch.nn.Module.__init__(self)

        # tokens
        self.token_to_id = deepcopy(token_to_id)
        self.padding_token = padding_token
        self.unknown_token_id = self.token_to_id[self.unknown_token]
        self.padding_token_id = self.token_to_id[self.padding_token]
        self.start_of_sequence_token = start_of_sequence_token
        self.start_of_sequence_token_id = self.token_to_id[start_of_sequence_token]

        self.id_to_token = {i: t for t, i in self.token_to_id.items()}

        # model hyperparameters
        self.embed_size = embed_size
        self.window_size = window_size
        self.vocab_size = vocab_size
        self.n_heads = n_heads
        self.n_blocks = n_blocks
        self.dropout = dropout
        self.activation = activation
        self.hidden_size = hidden_size

        # sampling
        self.sampling_temperature = sampling_temperature
        self._next_all_tokens = [
            self.id_to_token.get(i, self.unknown_token) for i in range(vocab_size)
        ]
        self.top_k = top_k
        self.top_p = top_p

        # learnable parameters
        self.token_embedding = torch.nn.Embedding(
            vocab_size, self.embed_size, padding_idx=self.padding_token_id
        )
        self.position_embedding = torch.nn.Embedding(self.window_size, self.embed_size)
        self.decoder_blocks = torch.nn.Sequential(
            *[
                DecoderBlock(
                    embed_size,
                    hidden_size,
                    n_heads,
                    max_seq_len=self.window_size,
                    dropout=dropout,
                    activation=activation,
                )
                for _ in range(n_blocks)
            ]
        )
        self.final_layer_norm = torch.nn.LayerNorm(self.embed_size)
        self.language_modeling_head = torch.nn.Linear(self.embed_size, self.vocab_size)
        # self.language_modeling_head = torch.nn.Linear(..., bias=False) # for tied weights

        # model configuration
        self.to(device)

        self.apply(self._initialize_weights)  # gaussian(mean=0, std=0.02)
        # apply special scaled initialization to the residual projections, per GPT-2 paper
        for parameter_name, param in self.named_parameters():
            if parameter_name.endswith("projection.weight"):
                torch.nn.init.normal_(
                    param, mean=0.0, std=0.02 / ((2 * n_blocks) ** 0.5)
                )

        if pretrained_token_embeddings is not None:
            assert (
                pretrained_token_embeddings.shape == self.token_embedding.weight.shape
            )
            self.token_embedding.load_state_dict(
                {"weight": pretrained_token_embeddings}
            )
            self.token_embedding.weight.requires_grad = False

        # weight tying (transfer learning becomes difficult)
        # self.language_modeling_head.weight = self.token_embedding.weight

        # TODO: experiment with weight tying with MLP of decoder
        # self.decoder_blocks[-1].feed_forward.fully_connected_1.weight = self.token_embedding.weight
        # self.decoder_blocks[-1].feed_forward.fully_connected_1.bias = 0
        # self.decoder_blocks[-1].feed_forward.fully_connected_1.bias.requires_grad = False

        # other
        self.n_parameters = sum(p.numel() for p in self.parameters())
        self.randomly_shift_position_embedding_during_training = (
            randomly_shift_position_embedding_during_training
        )
        self.training_history = {
            "train_epoch": [],
            "train_loss": [],
            "val_epoch": [],
            "val_loss": [],
        }
        self.eval()

    def forward(self, token_ids: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model.

        This method embeds the token_ids into vectors. It also embeds their positions into vectors.
        Depending upon the training & randomly_shift flags, it may shift sequences' position by a random amount.
        The embeddings are added together and passed to transformer decoder block containing causal multi-head self attention.
        The output is passed through LayerNorm and finally to a language-modeling-head which converts the vectors into logits for each token.

        Args:
            token_ids (torch.Tensor): Tensor containing the token IDs. Shape is ([batch,] time).

        Returns:
            torch.Tensor: Tensor containing the logits. Shape is ([batch,] time, vocab_size).
        """

        # token_ids.shape = ([batch,] time)
        token_ids = token_ids[..., -self.window_size :]
        position_ids = self._make_position_ids(token_ids)

        # embed the tokens
        token_embeddings = self.token_embedding(token_ids)  # (batch, time, embed_size)
        position_embeddings = self.position_embedding(
            position_ids
        )  # (time, embed_size)
        x = token_embeddings + position_embeddings

        # apply QKV & FF layers
        x = self.decoder_blocks(x)  # (batch, time, embed_size)

        # normalization
        x = self.final_layer_norm(x)  # (batch, time, embed_size)

        # TODO: predict target token's embedding & calculate loss using MSE
        # x = self.embedding_prediction(x)
        # TODO: add separately trained MLP as language_modeling_head

        # compute logits
        logits = self.language_modeling_head(x)  # ([batch,] time, vocab_size)

        return logits

    def next(self, context: Iterable) -> Tuple[Any, float]:
        next_tokens, probabilities = self.next_all(context)

        filtered_indexes = (
            np.arange(len(probabilities))
            if len(context) == 1 and context[0] == self.start_of_sequence_token  # type: ignore
            else top_p_top_k_indexes(probabilities, top_k=self.top_k, top_p=self.top_p)
        )

        index_in_filtered_indexes = sample_one_index(
            np.array(probabilities)[filtered_indexes].tolist(),  # type: ignore
            temperature=self.sampling_temperature,
        )

        next_token_index = filtered_indexes[index_in_filtered_indexes]
        return next_tokens[next_token_index], probabilities[next_token_index]  # type: ignore

    def next_all(self, context) -> Tuple[List[Any], List[float]]:
        context = context[-self.window_size :]  # type: ignore
        last_input_token_index = len(context) - 1

        ids = self.tokens_to_ids(context)  # type: ignore
        ids = torch.Tensor(ids).type(torch.long).to(self.device)

        with torch.no_grad():
            logits = self.forward(ids)
            logits = logits[..., last_input_token_index, :]
            probabilities = torch.nn.functional.softmax(logits, dim=-1)

        return (
            self._next_all_tokens,
            probabilities.tolist(),  # .squeeze()
        )

    @staticmethod
    def load(
        model_path, device="cuda" if torch.cuda.is_available() else "cpu"
    ) -> TransformerLanguageModel:
        """
        Loads a TransformerLanguageModel from a given model path.

        Args:
            model_path (str): The path to the saved model file.
            device (str, optional): The device to load the model on. Defaults to "cuda" if a CUDA device is available, else "cpu".

        Returns:
            TransformerLanguageModel: The loaded TransformerLanguageModel object.
        """

        model_state_dict: Dict[str, Any] = torch.load(
            model_path, map_location=torch.device(device)
        )
        model = TransformerLanguageModel(
            **{
                k: v
                for k, v in model_state_dict.items()
                if k not in ["state_dict", "training_history"]
            },
        )
        model.load_state_dict(model_state_dict["state_dict"])
        model.training_history.update(model_state_dict.get("training_history", {}))

        return model

    def save(self, model_path: str, overwrite: bool = False) -> None:
        """
        Save the model to a file.

        Args:
            model_path (str): The path to save the model.
            overwrite (bool, optional): Whether to overwrite an existing file. Defaults to False.

        Raises:
            FileExistsError: If there is already a file at the specified path and overwrite is set to False.
        """

        if exists(model_path) and not overwrite:
            raise FileExistsError(f"there is already a file at {model_path = }")

        torch.save(self.get_model_state(), model_path)

    def get_model_state(self) -> Dict[str, Any]:
        """Returns the current state of the model as a dictionary.

        Returns:
            Dict[str, Any]: A dictionary mapping strings to the class arguments,
                pytorch model's state_dict and other attributes.
        """

        return {
            # torch parameters
            "state_dict": self.state_dict(),
            # tokens
            "token_to_id": self.token_to_id,
            "padding_token": self.padding_token,
            "unknown_token": self.unknown_token,
            "start_of_sequence_token": self.start_of_sequence_token,
            # model hyperparameters
            "vocab_size": self.vocab_size,
            "window_size": self.window_size,
            "embed_size": self.embed_size,
            "hidden_size": self.hidden_size,
            "n_heads": self.n_heads,
            "n_blocks": self.n_blocks,
            "dropout": self.dropout,
            "activation": self.activation,
            # sampling
            "sampling_temperature": self.sampling_temperature,
            "top_k": self.top_k,
            "top_p": self.top_p,
            # other
            "name": self.name,
            "training_history": self.training_history,
            "randomly_shift_position_embedding_during_training": self.randomly_shift_position_embedding_during_training,
        }

    def tokens_to_ids(self, tokens: Iterable[str]) -> List[int]:
        """
        Convert a list of tokens into a list of corresponding token IDs.

        Args:
            tokens (Iterable[str]): A list of tokens.

        Returns:
            List[int]: A list of token IDs. If a token is not found in the token_to_id dictionary, the unknown_token_id is used instead.
        """

        return [self.token_to_id.get(token, self.unknown_token_id) for token in tokens]

    def ids_to_tokens(self, ids: Iterable[int] | torch.Tensor):
        """
        Convert a sequence of token IDs to tokens.

        Args:
            ids (Iterable[int] | torch.Tensor): An iterable of token IDs.

        Returns:
            List[str]: A list of tokens corresponding to the input IDs.
        """

        if isinstance(ids, int):
            ids = [ids]
        elif isinstance(ids, torch.Tensor):
            ids = ids.tolist()

        return [self.id_to_token.get(i, self.unknown_token) for i in ids]

    def to(self, device, *args, **kwargs):
        self.device = device
        super().to(device, *args, **kwargs)

        return self

    def _initialize_weights(self, module):
        if isinstance(module, torch.nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, torch.nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def _make_position_ids(self, token_ids: torch.Tensor) -> torch.Tensor:
        """Generates position IDs for the given token IDs (normally from 0 to seq_len-1).
        In case both model's training and randomly_shift_position_embedding_during_training attributes are True,
        the position IDs are shifted ahead ahead by a random amount according to the window_size
        except when the sequence starts with start_of_sequence token.

        Args:
            token_ids (torch.Tensor): The input token IDs.

        Returns:
            torch.Tensor: The generated position IDs.

        """
        seq_len = token_ids.shape[-1]
        position_ids = torch.arange(seq_len, device=self.device)

        # simulate that a substring can appear at any position inside the window
        # assuming train batches are bucketed by length (no extra padding in batch)
        # before: [0,1,2,3] -> [[0,1,2,3] _, _, _, ...] only initial positions get trained
        # after : [0,1,2,3] -> [..., _, _, _, [4,5,6,7] _, _, _, ...] all positions get trained
        if self.training and self.randomly_shift_position_embedding_during_training:
            batch_size = None if token_ids.dim() == 1 else token_ids.shape[0]
            shifts = torch.randint(
                0,
                self.window_size - seq_len + 1,
                (batch_size, 1) if batch_size else (1,),
                device=self.device,
            )
            # dont shift examples that start with SOS token
            shifts[token_ids[..., :1] == self.start_of_sequence_token_id] = 0

            position_ids = position_ids + shifts

        return position_ids

    def __str__(self) -> str:
        return f"Transformer LM: {super().__str__()}, window={self.window_size}, params={self.n_parameters}"


__all__ = [
    "TransformerLanguageModel",
]



"""
sign_language_translator
========================
Code: https://github.com/sign-language-translator/sign-language-translator
Help: https://sign-language-translator.readthedocs.io

This project is an effort to bridge the communication gap between the hearing and the hearing-impaired community using Artificial Intelligence.
The goal is to provide a user friendly API to novel Sign Language Translation solutions that can easily adapt to any regional sign language.

Usage
-----

.. code-block:: python

    import sign_language_translator as slt

    # download dataset or models (if you need them for personal use)
    # (by default, resources are auto-downloaded within the install directory)
    # slt.Assets.set_root_dir("path/to/folder")  # Helps preventing duplication across environments or using cloud synced data
    # slt.Assets.download(".*.json")  # downloads into resource_dir
    # print(slt.Settings.FILE_TO_URL.keys())  # All downloadable resources

    print("All available models:")
    print(list(slt.ModelCodes))  # slt.ModelCodeGroups
    # print(list(slt.TextLanguageCodes))
    # print(list(slt.SignLanguageCodes))
    # print(list(slt.SignFormatCodes))

    # -------------------------- TRANSLATE: text to sign --------------------------

    import sign_language_translator as slt

    # Load text-to-sign model
    # deep_t2s_model = slt.get_model("t2s-flan-T5-base-01.pt") # pytorch

    # rule-based model (concatenates clips of each word)
    t2s_model = slt.models.ConcatenativeSynthesis(
        text_language = "urdu", # or object of any child of slt.languages.text.text_language.TextLanguage class
        sign_language = "pakistan-sign-language", # or object of any child of slt.languages.sign.sign_language.SignLanguage class
        sign_format = "video", # or object of any child of slt.vision.sign.Sign class
    )

    text = "HELLO دنیا!" # HELLO treated as an acronym
    sign_language_sentence = t2s_model(text)

    # sign_language_sentence.show() # class: slt.vision.sign.Sign or its child
    # sign_language_sentence.save(f"sentences/{text}.mp4")

    # -------------------------- TRANSLATE: sign to text --------------------------

    import sign_language_translator as slt

    # # Load sign-to-text model (pytorch) (COMING SOON!)
    # translation_model = slt.get_model(slt.ModelCodes.Gesture)
    embedding_model = slt.models.MediaPipeLandmarksModel()

    sign = slt.Video("video.mp4")
    embedding = embedding_model.embed(sign.iter_frames())
    # text = translation_model.translate(embedding)

    # print(text)
    sign.show()
    # slt.Landmarks(embedding, connections="mediapipe-world").show()
"""

from sign_language_translator import config, languages, models, text, utils, vision
from sign_language_translator.config import enums
from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import ModelCodeGroups, ModelCodes
from sign_language_translator.config.enums import SignFormats as SignFormatCodes
from sign_language_translator.config.enums import SignLanguages as SignLanguageCodes
from sign_language_translator.config.enums import TextLanguages as TextLanguageCodes
from sign_language_translator.config.settings import Settings
from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.models import get_model
from sign_language_translator.vision._utils import get_sign_wrapper_class
from sign_language_translator.vision.video.video import Video

__version__ = config.utils.get_package_version()

__all__ = [
    # config
    "Assets",
    "Settings",
    "__version__",
    # modules
    "vision",
    "text",
    "models",
    "languages",
    "utils",
    "config",
    "enums",
    # classes (enum)
    "ModelCodes",
    "TextLanguageCodes",
    "SignLanguageCodes",
    "SignFormatCodes",
    "ModelCodeGroups",
    # classes (wrappers)
    "Video",
    # object loaders
    "get_sign_language",
    "get_text_language",
    "get_model",
    "get_sign_wrapper_class",
]



"""This module provides utility functions for video processing.
"""

from mimetypes import guess_type
from typing import Generator, List, Sequence, Tuple

import cv2
from numpy import uint8
from numpy.typing import NDArray

__all__ = [
    "read_frames_with_opencv",
    "iter_frames_with_opencv",
]


def read_frames_with_opencv(path: str) -> List[NDArray[uint8]]:
    """
    Extracts individual frames from a video file or an image file.

    This function reads a video file using opencv and extracts its frames as numpy arrays.
    It can also read an image file and treat it as a single frame video.

    Args:
        path (str): The path to the video or image file.

    Returns:
        List[NDArray]: A list of numpy arrays, each representing a frame from the video.

    Raises:
        FileNotFoundError: If the video file is not found or cannot be opened.
    """

    return list(iter_frames_with_opencv(path))


def iter_frames_with_opencv(path: str) -> Generator[NDArray[uint8], None, None]:
    """
    Extracts individual frames from a video file or an image file.

    This function reads a video file using opencv and extracts its frames as numpy arrays.
    It can also read an image file and treat it as a single frame video.

    Args:
        path (str): The path to the video or image file.

    Returns:
        Generator[NDArray, None, None]: yields numpy arrays representing frames from the video (W, H, C).

    Raises:
        FileNotFoundError: If the video file is not found or cannot be opened.
    """

    file_type = str(guess_type(path)[0])
    if not file_type.startswith(("image", "video")):
        raise ValueError(f"unknown file type: {file_type}")

    if file_type.startswith("image"):
        frame = cv2.imread(path)
        yield cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)  # type: ignore

    elif file_type.startswith("video"):
        capture = cv2.VideoCapture(path)
        for _ in range(int(capture.get(cv2.CAP_PROP_FRAME_COUNT))):
            ret, frame = capture.read()
            if not ret:
                break

            yield cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)  # type: ignore

        capture.release()


def _normalize_args_index_and_timestamp(
    timestamp: float | None, index: int | None, max_duration: float, max_index: int
) -> Tuple[float, int]:
    if (timestamp is not None) and index is None:
        if not 0 <= timestamp <= max_duration:
            raise ValueError(
                f"timestamp must be between 0 and {max_duration}, but got {timestamp}"
            )

        return timestamp, round(
            (timestamp / max_duration if max_duration else 1) * max_index
        )

    elif (index is not None) and timestamp is None:
        if not 0 <= index <= max_index:
            raise ValueError(
                f"index must be between 0 and {max_index}, but got {index}."
            )

        return index / (max_index or 1) * max_duration, index

    else:
        raise ValueError("provide either timestamp or index.")


def _validate_and_normalize_slices(
    keys: int | slice | Sequence[int | slice], max_n_dims: int = 4
) -> Tuple[slice, ...]:
    if not isinstance(keys, Sequence):
        keys = [keys]

    slices = []
    for i, key in enumerate(keys):
        if key is Ellipsis:
            slices += [slice(None)] * (max_n_dims - len(keys) + 1)
        elif isinstance(key, int):
            slices.append(slice(key, key + 1))
        elif isinstance(key, slice):
            slices.append(key)
        else:
            raise ValueError(
                f"Invalid argument: {key} at index {i}. Provide either an integer, slice or ellipsis."
            )

    return tuple(slices)



"""load word datasets to create word maps etc."""

import json
import os
import re
from typing import Any, Dict, Iterable, List, Set, TypedDict

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.settings import Settings

__all__ = [
    "Vocab",
    "MappingDataset",
]


# =================================== #
#    Mapping Dataset Static Typing    #
# =================================== #


class Mapping(TypedDict):
    components: List[str]
    """list of sign videos that can make up the video identified by label"""
    label: str
    """the filename of the sign video"""
    token: Dict[str, List[str]]
    """maps language codes to list of tokens that correspond to the label"""
    gloss: Dict[str, List[str]]
    """maps language codes to list of glosses that correspond to the video. A gloss is a word-for-word translation of the signs in the video."""
    translation: Dict[str, List[str]]
    """maps language codes to list of translations that correspond to the video. A translation is a grammatically correct text that has the same meaning as the sign sequence in the video."""


class MappingDataset(TypedDict):
    country: str
    description: str
    mapping: List[Mapping]
    organization: str
    url: str


# ================= #
#    Data Reader    #
# ================= #


class Vocab:
    """Loads text datasets for a specific language, country and organization.

    Note:
        Our mapping datasets will only be downloaded automatically if the `data_root_dir` arg is the same as `Assets.ROOT_DIR`.
    """

    def __init__(
        self,
        language: str = r".^",
        country: str = r".^",
        organization: str = r".^",
        part_number: str = r".^",
        data_root_dir: str = Assets.ROOT_DIR,
        arg_is_regex: bool = True,
        word_sense_regex: str = r"\([^\(\)]*\)",
    ) -> None:
        # save arguments
        self.language = language
        self.country = country
        self.organization = organization
        self.part_number = part_number
        self.data_root_dir = os.path.abspath(data_root_dir)
        self.arg_is_regex = arg_is_regex
        self.word_sense_regex = word_sense_regex

        # initialize properties with defaults
        self.word_to_labels: Dict[str, List[List[str]]] = {}
        self.supported_words_with_word_sense: Set[str] = set()
        self.supported_words: Set[str] = set()  # TODO: why does this variable exist?
        self.ambiguous_to_unambiguous: Dict[str, List[str]] = {}
        self.person_names: List[str] = []
        self.words_to_numbers: Dict[str, int] = {}
        self.misspelled_to_correct: Dict[str, str] = {}
        self.number_suffix_to_zeros: Dict[str, str] = {}
        self.joint_word_to_split_words: Dict[str, str] = {}
        self.numeric_keys: Set[str] = set()

        # load data
        self.__load_mapping_datasets()
        self.__load_preprocessing()

    def remove_word_sense(self, text: str) -> str:
        """Remove the word sense or disambiguation information from given text.

        Args:
            text (str): The text from which the word sense needs to be removed.

        Returns:
            str: The word without the word sense or disambiguation information.

        Example:
            word = "this is a spring(metal-coil). those are glasses(water-containers)."
            without_word_sense = remove_word_sense(word)
            print(without_word_sense)  # Output: "this is a spring. those are glasses."
        """

        without_word_sense = re.sub(self.word_sense_regex, "", text)
        return without_word_sense

    def __load_preprocessing(self) -> None:
        self.__download_resource(fname := "text-preprocessing.json")
        with open(os.path.join(self.data_root_dir, fname), "r", encoding="utf-8") as f:
            raw_data: Dict[str, Dict[str, Any]] = json.load(f)

        data: Dict[str, Any] = {
            key: lang_to_data.get(lang, self.__default_value(lang_to_data))
            for key, lang_to_data in raw_data.items()
            for lang in lang_to_data
            if self.__match(self.language, lang, self.arg_is_regex)
        }

        self.person_names: List[str] = data.get("person_names", [])
        self.words_to_numbers: Dict[str, int] = data.get("words_to_numbers", {})
        self.misspelled_to_correct: Dict[str, str] = data.get(
            "misspelled_to_correct", {}
        )
        self.number_suffix_to_zeros: Dict[str, str] = data.get(
            "number_suffixes_to_zeros", {}
        )
        self.joint_word_to_split_words: Dict[str, str] = data.get(
            "joint_word_to_split_words", {}
        )
        # TODO: improve coverage of key.isnumeric()
        self.numeric_keys: Set[str] = {
            key for key in self.word_to_labels if key.isnumeric()
        }

    def __load_mapping_datasets(self) -> None:
        # download conditionally
        self.__download_resource(
            filename := f"{self.country.rstrip('$')}-dictionary-mapping.json"
        )

        # load existing
        mapping_filepaths = [
            os.path.join(self.data_root_dir, file)
            for file in os.listdir(self.data_root_dir)
            if re.match(filename, file)
        ]
        for filepath in mapping_filepaths:
            with open(filepath, "r", encoding="utf-8") as f:
                self.word_to_labels.update(
                    self._make_word_to_labels(
                        self.language,
                        self.country,
                        self.organization,
                        self.part_number,
                        json.load(f),
                    )
                )

        self.supported_words_with_word_sense = set(self.word_to_labels)
        self.supported_words = {
            self.remove_word_sense(w) for w in self.supported_words_with_word_sense
        }
        self.ambiguous_to_unambiguous = self._make_disambiguation_map(
            self.supported_words_with_word_sense
        )
        self.labels: Set[str] = {
            label
            for label_sequences in self.word_to_labels.values()
            for label_sequence in label_sequences
            for label in label_sequence
        }

    def _make_word_to_labels(
        self,
        language: str,
        country: str,
        organization: str,
        part_number: str,
        mapping_datasets: List[MappingDataset],
        is_regex: bool = True,
    ) -> Dict[str, List[List[str]]]:
        """
        Takes JSON word mapping datasets and creates a dictionary mapping text tokens to sign labels.

        Args:
            language (str | None, optional): Language code used in JSON or a regex matching it whose data should be extracted. Defaults to None.
            country (str): Country code used to filter the mapping datasets.
            organization (str): Organization code used to filter the mapping datasets.
            part_number (str): Part number used to filter the mapping datasets.
            mapping_datasets (List[MappingDataset]): List of mapping datasets.
            is_regex (bool, optional): Treat the provided language code, country, organization and part_number as regex when comparing against JSON keys. Defaults to True.

        Returns:
            Dict[str, List[List[str]]]: A dictionary that maps each text token to a list containing sequences of signs.
        """

        word_to_labels: Dict[str, List[List[str]]] = {}

        for dataset in mapping_datasets:
            # different country or organization
            if not (
                self.__match(country, dataset["country"], is_regex)
                and self.__match(organization, dataset["organization"], is_regex)
            ):
                continue

            for mapping in dataset["mapping"]:
                # different part number
                if not all(
                    self.__match(
                        part_number, label.split("_")[0].split("-")[-1], is_regex
                    )
                    for label in ([mapping["label"]] if "label" in mapping else [])
                    + mapping.get("components", [])
                ):
                    continue

                for lang, token_list in mapping.get("token", {}).items():
                    # different language
                    if not self.__match(language, lang, is_regex):
                        continue

                    for token in token_list:
                        if token not in word_to_labels:
                            word_to_labels[token] = []
                        if "label" in mapping:
                            word_to_labels[token].append([mapping["label"]])
                        if "components" in mapping:
                            # ???? Why filter down the components??????
                            if all(
                                self.__match(country, x[0], is_regex)
                                and self.__match(organization, x[1], is_regex)
                                for comp in mapping["components"]
                                if (
                                    x := comp.split(Settings.FILENAME_SEPARATOR)[
                                        0
                                    ].split(Settings.FILENAME_CONNECTOR)
                                )
                            ):
                                word_to_labels[token].append(mapping["components"])

        # Drop duplicates (when language = r".*", digits, etc. are repeated)
        word_to_labels = {
            word: list(list(seq) for seq in set(tuple(seq) for seq in label_seqs))
            for word, label_seqs in word_to_labels.items()
        }

        return word_to_labels

    def _make_disambiguation_map(self, words: Iterable[str]) -> Dict[str, List[str]]:
        """create a mapping from ambiguous words to possible unambiguous words

        Args:
            words (Iterable[str]): A list of disambiguated words. A disambiguated word is a word that has a word-sense in it. e.g. ["spring(season)", "spring(bouncy-coil)"].

        Returns:
            Dict[str, List[str]]: A dictionary mapping ambiguous words to possible unambiguous words. e.g. {"spring": ["spring(season)", "spring(bouncy-coil)"]}.
        """
        ambiguous_2_unambiguous = {}
        for word in words:
            without_word_sense = self.remove_word_sense(word)
            if without_word_sense != word:
                if without_word_sense not in ambiguous_2_unambiguous:
                    ambiguous_2_unambiguous[without_word_sense] = []
                ambiguous_2_unambiguous[without_word_sense].append(word)

        return ambiguous_2_unambiguous

    def __match(self, text_1: str, text_2: str, text_1_is_regex: bool) -> bool:
        return bool(re.match(text_1, text_2)) if text_1_is_regex else text_1 == text_2

    def __default_value(self, dictionary: Dict) -> Any:
        """get an empty instance of dict's first value's class"""
        first_value = next(iter(dictionary.values()), None)
        value_class = type(first_value)
        default_obj = value_class()

        return default_obj

    def __download_resource(self, file_name: str):
        if (
            Settings.AUTO_DOWNLOAD
            and not os.path.exists(os.path.join(self.data_root_dir, file_name))
            and self.data_root_dir == Assets.ROOT_DIR
        ):
            Assets.download(file_name, overwrite=False)



"""
sign_language_translator.models
===============================

This module contains the various models in the sign language translator system
and their associated components.

Language Models:
----------------
- NgramLanguageModel: A language model based on n-grams.
- TransformerLanguageModel: A transformer-based language model.
- MixerLM: A language model that combines multiple language models using mixing weights.
- LanguageModel: An abstract base class for all language models in this package.
- BeamSampling: A utility class that performs beam search during text generation.

Text to Sign Translation:
-------------------------
- TextToSignModel: An abstract base class for all model that translates text into sign language gestures in this package.
- ConcatenativeSynthesis: A rule-based model for synthesizing sign language gestures from text.

Video Embedding:
----------------
- VideoEmbeddingModel: An abstract model that embeds video frames into a vector space.
- MediaPipeLandmarksModel: A video embedding model that utilizes MediaPipe for pose and hand landmark extraction.

Utilities:
----------
- get_model: A utility function to get any model by string name.
- utils: Miscellaneous utility functions for the sign language translator system.
"""

from sign_language_translator.models import (
    language_models,
    sign_to_text,
    text_to_sign,
    utils,
    video_embedding,
)
from sign_language_translator.models._utils import get_model
from sign_language_translator.models.language_models import (
    BeamSampling,
    LanguageModel,
    MixerLM,
    NgramLanguageModel,
    TransformerLanguageModel,
)
from sign_language_translator.models.text_to_sign import ConcatenativeSynthesis
from sign_language_translator.models.text_to_sign.t2s_model import TextToSignModel
from sign_language_translator.models.video_embedding import (
    MediaPipeLandmarksModel,
    VideoEmbeddingModel,
)

__all__ = [
    "get_model",
    "language_models",
    "sign_to_text",
    "text_to_sign",
    "utils",
    "video_embedding",
    "ConcatenativeSynthesis",
    "NgramLanguageModel",
    "TransformerLanguageModel",
    "BeamSampling",
    "MixerLM",
    "LanguageModel",
    "TextToSignModel",
    "MediaPipeLandmarksModel",
    "VideoEmbeddingModel",
]



"""defines settings constants for global context

Module Structure:
- Settings (class): contains config constants as class variables.
"""

__all__ = [
    "Settings",
]


class Settings:
    """Class containing settings and configuration parameters
    for the sign language translator library."""

    FILENAME_SEPARATOR = "_"
    """The separator used in dataset filenames to separate different attributes."""

    FILENAME_CONNECTOR = "-"
    """The connector used in filenames to join parts of same attribute."""

    AUTO_DOWNLOAD = True
    """A flag indicating whether automatic downloading of missing dataset files is enabled."""

    SHOW_DOWNLOAD_PROGRESS = True
    """A flag indicating whether the download progress should be shown by default."""



"""
This module provides a flexible and unified interface for working with video frames.

This module defines an abstract class and its implementations for efficient video frame retrieval.
It includes various classes for accessing video frames, seeking, and caching from various sources.

Classes:
- VideoFrames: An abstract base class for video frame retrieval.
- VideoCaptureFrames: A class for efficient video frame retrieval from a video file using OpenCV.
- SequenceFrames: A class for representing a sequence of video frames.
- IterableFrames: Represents an iterable video frame source, allowing random access to frames by index or timestamp.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from os.path import abspath
from time import time
from typing import Iterable, Sequence

import cv2
import numpy as np
from numpy.typing import NDArray

from sign_language_translator.vision.utils import _normalize_args_index_and_timestamp


class VideoFrames(ABC):
    """
    Abstract Base Class for Video Frames

    VideoFrames is an abstract base class that defines a common interface for video frame retrieval.
    Subclasses of VideoFrames are expected to implement methods for accessing video frames,
    releasing resources, and providing information about the video.

    Methods:
    - get_frame(timestamp: float | None = None, index: int | None = None) -> NDArray[np.uint8]:
        Get a frame at a given timestamp or index from the video object.

    - close():
        Release the resources occupied by the object.

    - __len__() -> int:
        Return the number of frames in the video object.

    Properties:
    - height: int
        Number of pixels vertically present in the video frame.

    - width: int
        Number of pixels horizontally present in the video frame.

    - n_channels: int
        Number of color channels in the video frames.
    """

    @abstractmethod
    def get_frame(
        self, timestamp: float | None = None, index: int | None = None
    ) -> NDArray[np.uint8]:
        """Get a frame at a given timestamp or index from the video object."""

    @abstractmethod
    def close(self):
        """Release the resources occupied by the object."""

    @abstractmethod
    def __len__(self) -> int:
        """Number of frames in the video object"""

    @property
    @abstractmethod
    def height(self) -> int:
        """Number of pixels vertically present in the video frame."""

    @property
    @abstractmethod
    def width(self) -> int:
        """Number of pixels horizontally present in the video frame."""

    @property
    @abstractmethod
    def n_channels(self) -> int:
        """Number of color channels in the video frames."""


class VideoCaptureFrames(VideoFrames):
    """
    A class for efficient video frame retrieval from a video file using OpenCV.

    This class extends the functionality of the VideoFrames abstract class to provide features
    for video frame access, seeking, and caching.

    Args:
        path (str): The path to the video file.
        use_cache (bool, optional): Enable or disable frame caching. Default is False.
        cache_len (int, optional): Maximum number of frames to cache if use_cache is enabled.
            Default is 256.
        **kwargs: Additional keyword arguments to pass to the base VideoFrames class.

    Attributes:
        path (str): The path to the video file.
        fps (float): Frames per second of the video.
        total_frames (int): Total number of frames in the video.
        _width (int): Width of video frames.
        _height (int): Height of video frames.
        fourcc (int): FourCC code representing the video codec.
        duration (float): Duration of the video in seconds.
        _frames_cache (dict): A dictionary for frame caching.
        use_cache (bool): True if frame caching is enabled, False otherwise.
        _max_cache_len (int): Maximum number of frames to cache.
        _n_channels (int): Number of color channels in the video frames.

    Methods:
        get_frame(timestamp: float | None = None, index: int | None = None) -> NDArray[np.uint8]:
            Retrieve a video frame based on either a timestamp or an index.

        current_index() -> int:
            Get the current index of the video frame being read.

        seek(timestamp: float | None = None, index: int | None = None):
            Move the video frame position to the specified timestamp or index.

        read_frame() -> NDArray[np.uint8] | None:
            Read and return the next frame from the video.

        close():
            Close the video capture and release associated resources.

    Notes:
        - Frame caching can improve performance by storing previously accessed frames in memory.
        - The seek method employs efficient seeking techniques based on time and frame index.
        - When finished, remember to call the close method to release video resources.

    Example usage:
        video = VideoCaptureFrames("video.mp4", use_cache=True)
        frame = video.get_frame(timestamp=10.0)
        video.seek(index=100)
        frame = video.read_frame()
        video.close()
    """

    _seek_time: float = 0.1
    _read_time: float = 0.005

    def __init__(self, path: str, use_cache=False, cache_len=256, **kwargs):
        self.path = abspath(path)
        self.video_capture = cv2.VideoCapture(self.path)

        self.fps = float(self.video_capture.get(cv2.CAP_PROP_FPS)) or 30
        self.total_frames = int(
            self.video_capture.get(cv2.CAP_PROP_FRAME_COUNT)
        )  # bug: not accurate, gotta fix on runtime.
        self._width = int(self.video_capture.get(cv2.CAP_PROP_FRAME_WIDTH))
        self._height = int(self.video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.fourcc = int(self.video_capture.get(cv2.CAP_PROP_FOURCC))

        self._frames_cache = {}
        self.use_cache = use_cache
        self._max_cache_len = cache_len

        frame_shape = self.get_frame(0).shape
        self._n_channels = frame_shape[2] if len(frame_shape) == 3 else 1

    def get_frame(
        self,
        timestamp: float | None = None,
        index: int | None = None,
    ) -> NDArray[np.uint8]:
        """
        Retrieve a video frame at a specified timestamp or index.

        Args:
            timestamp (float | None): The timestamp in seconds.
            index (int | None): The frame index.

        Returns:
            NDArray[np.uint8]: The video frame as a NumPy array.

        Raises:
            RuntimeError: If frame retrieval fails.
        """

        # arguments
        timestamp, _index = _normalize_args_index_and_timestamp(
            timestamp, index, self.duration, self.total_frames - 1
        )

        # cache
        if self.use_cache and _index in self._frames_cache:
            return self._frames_cache[_index]

        # read
        self.seek(index=_index)
        frame = self.read_frame()

        # validate
        if frame is None:
            raise RuntimeError(
                f'Error reading frame from cv2.VideoCapture("{self.path}") at '
                + (f"{timestamp = }." if index is None else f"{index = }.")
            )

        # cache
        if self.use_cache:
            if len(self._frames_cache) > self._max_cache_len:
                self._frames_cache.pop(list(self._frames_cache.keys())[0])

            self._frames_cache[_index] = frame

        return frame

    @property
    def current_index(self) -> int:
        """Where the VideoCapture is currently pointing to."""
        return int(self.video_capture.get(cv2.CAP_PROP_POS_FRAMES))

    def seek(self, timestamp: float | None = None, index: int | None = None):
        """
        Seek to a specified timestamp or frame index.

        Args:
            timestamp (float | None): The timestamp in seconds.
            index (int | None): The frame index.

        Returns:
            None
        """
        timestamp, target_index = _normalize_args_index_and_timestamp(
            timestamp, index, self.duration, self.total_frames - 1
        )

        if target_index == self.current_index:
            return

        new_seek_time = None
        new_read_time = None
        max_frames_to_read = int(self._seek_time / self._read_time * 0.9)

        if self.current_index < target_index < self.current_index + max_frames_to_read:
            # go forward by reading frames because it is faster than seeking
            n_frames = target_index - self.current_index
            start_time = time()
            for _ in range(n_frames):
                self.read_frame()
            new_read_time = (time() - start_time) / n_frames
        else:
            # seek using cv2
            start_time = time()
            self.video_capture.set(cv2.CAP_PROP_POS_FRAMES, target_index)
            new_seek_time = time() - start_time

        # updated seek and read times using exponential moving average
        if new_seek_time is not None:
            self._seek_time = self._seek_time * 0.90 + 0.10 * new_seek_time
        if new_read_time is not None:
            self._read_time = self._read_time * 0.95 + 0.05 * new_read_time

    def read_frame(self) -> NDArray[np.uint8] | None:
        """
        Read the next frame from the video.

        Returns:
            NDArray[np.uint8] | None: The next video frame as a NumPy array,
                or None if no more frames are available.
        """
        success, frame = self.video_capture.read()
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB) if success else None

        return frame  # type: ignore

    def __len__(self) -> int:
        return self.total_frames

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width

    @property
    def n_channels(self) -> int:
        return self._n_channels

    @property
    def duration(self) -> float:
        return self.total_frames / self.fps

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """Release the video capture resource and clear the frame cache."""
        self.video_capture.release()
        self._frames_cache = {}

    def __copy__(self) -> VideoCaptureFrames:
        new = self.__class__.__new__(self.__class__)
        new.__dict__.update(self.__dict__)
        # TODO: decide if create new videoCapture to prevent conflicts or point to old because this is shallow copy
        # new.video_capture = self.video_capture
        new.video_capture = cv2.VideoCapture(self.path)
        return new

    def __deepcopy__(self, memo):
        new_instance = self.__class__(
            self.path,
            use_cache=self.use_cache,
            cache_len=self._max_cache_len,
        )
        memo[id(self)] = new_instance
        return new_instance

    # TODO: make the class pickle-able


class SequenceFrames(VideoFrames):
    """A class for representing a sequence of video frames.

    This class extends the VideoFrames abstract class to work with a predefined sequence
    of frames, allowing easy access to individual frames within the sequence.

    Args:
        frames (Sequence[NDArray[np.uint8]]): A sequence of video frames, where each
            frame is represented as a NumPy array with data type `np.uint8`.
        fps (float | None, optional): The frames per second (FPS) of the video.
            If not specified, it can be set to None.

    Attributes:
        frames (Sequence[NDArray[np.uint8]]): The sequence of video frames.
        fps (float): The frames per second (FPS) of the video. Defaults to 30.0.

    Note:
        The SequenceFrames class inherits from the VideoFrames class.
    """

    def __init__(
        self,
        frames: Sequence[NDArray[np.uint8]],
        fps: float = 30.0,
    ) -> None:
        self.frames = frames
        self.fps = abs(fps) or 30.0

        frame_shape = self.frames[0].shape
        self._height, self._width = frame_shape[:2]
        self._n_channels = frame_shape[2] if len(frame_shape) == 3 else 1

    def get_frame(
        self, timestamp: float | None = None, index: int | None = None
    ) -> NDArray[np.uint8]:
        """
        Retrieve a video frame based on the specified timestamp or index.

        Args:
            timestamp (float | None, optional): The timestamp in seconds at which to
                retrieve the frame. If not provided, index is used.
            index (int | None, optional): The index of the frame to retrieve. If not
                provided, timestamp is used.

        Returns:
            NDArray[np.uint8]: The video frame as a NumPy array with data type `np.uint8`.
        """

        timestamp, _index = _normalize_args_index_and_timestamp(
            timestamp, index, self.duration, self.total_frames - 1
        )
        return self.frames[_index]

    def close(self):
        """
        Close the SequenceFrames instance by clearing the frames.

        This method releases resources associated with the frames by clearing the
        frames list. After calling this method, the frames will no longer be available.
        """
        self.frames = []

    def __len__(self) -> int:
        return self.total_frames

    @property
    def total_frames(self) -> int:
        """total number of frames present in the sequence. (dimension=0)"""
        return len(self.frames)

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width

    @property
    def n_channels(self) -> int:
        return self._n_channels

    @property
    def duration(self) -> float:
        """total time that the frames would take to play in a sequence. depends on fps."""
        return (self.total_frames / self.fps) if self.fps else float("inf")


class IterableFrames(VideoFrames):
    """
    Represents a read-once iterable video frame source, allowing random access to frames by index or timestamp.

    This class extends the VideoFrames abstract class and is specifically designed to work with iterable sources of video frames such as generators.
    It maintains an internal cache of frames to efficiently access frames by index or timestamp.

    Args:
        frames (Iterable[NDArray[np.uint8]]): An iterable source of video frames.
        total_frames (int): The total number of frames in the video.
        fps (float, optional): The frames per second of the video. Defaults to 30.0.
        use_cache (bool, optional): Whether to store the frames when they have been read from the iterable. Defaults to True.

    Attributes:
        frames_iterable (iter): An iterator over the provided frames.
        frames_cache (dict): A cache to store frames for efficient retrieval.
        fps (float): The frames per second of the video.
        total_frames (int): The total number of frames in the video.

    Methods:
        get_frame(timestamp: float | None = None, index: int | None = None) -> NDArray[np.uint8]:
            Retrieve a video frame by specifying either a timestamp or an index.

        close():
            Close the video frame source, resetting the frames_iterable and clearing the frames_cache.
    """

    def __init__(
        self,
        frames: Iterable[NDArray[np.uint8]],
        total_frames: int,
        fps: float = 30.0,
        use_cache=True,
    ) -> None:
        self.frames_iterable = iter(frames)
        self.frames_cache = {}
        self.fps = abs(fps) or 30.0
        self.use_cache = use_cache

        if (not isinstance(total_frames, int)) or total_frames <= 0:
            raise ValueError("total_frames must be a positive integer.")

        self.total_frames = total_frames

        self._current_index = 0

        frame = self.get_frame(index=0)
        self._height, self._width = frame.shape[:2]
        self._n_channels = frame.shape[2] if len(frame.shape) == 3 else 1

    def get_frame(
        self, timestamp: float | None = None, index: int | None = None
    ) -> NDArray[np.uint8]:
        """
        Retrieve a video frame by specifying either a timestamp or an index.

        Args:
            timestamp (float | None, optional): The timestamp (in seconds) of the desired frame. If provided, it will
                be used to locate the frame in the video. Defaults to None.
            index (int | None, optional): The index of the desired frame. If provided, it will be used to locate the
                frame in the video. Defaults to None.

        Returns:
            NDArray[np.uint8]: The video frame as a NumPy array of unsigned 8-bit integers.

        Raises:
            RuntimeError: If the specified timestamp or index is out of range or if there is an error reading the frame
                from the frames_iterable.

        Note:
            You can retrieve frames by either timestamp or index. The timestamp allows you to seek to a specific point
            in time, while the index allows you to access frames in a sequential order.
        """

        timestamp, target_index = _normalize_args_index_and_timestamp(
            timestamp, index, self.duration, self.total_frames - 1
        )

        for i in range(self._current_index, target_index + 1):
            frame = next(self.frames_iterable, None)
            if frame is not None:
                self.frames_cache[i] = frame
                self._current_index += 1
                continue

            # end of Video
            self.total_frames = self._current_index
            raise RuntimeError(
                "Error reading frame from frames_iterable at "
                + (f"{timestamp = }." if index is None else f"{index = }.")
            )

        return (
            self.frames_cache[target_index]
            if self.use_cache or target_index == 0
            else self.frames_cache.pop(target_index)
        )

    def close(self):
        """Close the video frame source, resetting the frames_iterable and clearing the frames_cache."""
        self.frames_iterable = iter([])
        self.frames_cache = {}

    def __len__(self) -> int:
        return self.total_frames

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width

    @property
    def n_channels(self) -> int:
        return self._n_channels

    @property
    def duration(self) -> float:
        """total time that the frames would take to play in a sequence. depends on fps."""
        return (self.total_frames / self.fps) if self.fps else float("inf")

    def __copy__(self) -> IterableFrames:
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result


import math
from typing import Callable, List


class VideoSource(VideoFrames):
    def __init__(
        self,
        source: VideoFrames,
        start_index: int = 0,
        end_index: int | None = None,
        step_size: int = 1,
        transformations: List[Callable] | None = None,
    ) -> None:
        self.source = source
        if start_index < 0 or start_index >= len(self.source):
            raise ValueError(
                "start_index cannot be negative or larger than the length of the source."
            )
        self.start_index = start_index
        self.end_index = end_index or len(source) - 1
        if self.end_index < start_index or self.end_index >= len(self.source):
            raise ValueError(
                "end_index cannot be negative, less than start_index or larger than the length of the source."
            )
        self._step_size = step_size or 1
        self.transformations = transformations or []

        self.__current_index = 0

    def __iter__(self):
        self.__current_index = 0
        return self

    def __next__(self) -> NDArray[np.uint8]:
        if 0 <= self.__current_index < len(self):
            frame = self.get_frame(index=self.__current_index)
            self.__current_index += 1
            return frame

        raise StopIteration

    @property
    def step_size(self) -> int:
        return self._step_size

    @step_size.setter
    def step_size(self, value: int):
        if not isinstance(value, int):
            raise ValueError("step_size must be an integer.")
        if value > 0:
            self.end_index -= (self.end_index - self.start_index) % value
        elif value < 0:
            self.start_index -= (self.end_index - self.start_index) % value
        else:
            raise ValueError("step_size cannot be 0")

        self._step_size = value

    def base_index(self, relative_index: int) -> int:
        return (
            self.start_index + (relative_index * self.step_size)
            if self.step_size > 0
            else self.end_index + (relative_index * self.step_size)
        )

    def get_frame(
        self, timestamp: float | None = None, index: int | None = None
    ) -> NDArray[np.uint8]:
        timestamp, index = _normalize_args_index_and_timestamp(
            timestamp,
            index,
            getattr(self.source, "duration", float("inf")),
            len(self) - 1,
        )

        if (base_index := self.base_index(index)) <= self.end_index:
            frame = self.source.get_frame(index=base_index)
            for transformation in self.transformations:
                frame = transformation(frame)
            return frame

        raise IndexError(
            "Error reading frame at "
            + (f"{timestamp = }." if index is None else f"{index = }.")
        )

    def close(self):
        self.source.close()

    def __len__(self) -> int:
        return math.ceil((self.end_index - self.start_index + 1) / abs(self.step_size))

    @property
    def height(self) -> int:
        return self.source.height

    @property
    def width(self) -> int:
        return self.source.width

    @property
    def n_channels(self) -> int:
        return self.source.n_channels
